#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import overload

from pydantic import PrivateAttr

from synnax import framer
from synnax.channel.payload import (
    Key,
    Operation,
    Params,
    Payload,
    normalize_params,
    ontology_id,
)
from synnax.channel.retrieve import Retriever
from synnax.channel.writer import Writer
from synnax.exceptions import MultipleFoundError, NotFoundError, ValidationError
from synnax.ontology.payload import ID
from synnax.telem import (
    CrudeDataType,
    CrudeSeries,
    CrudeTimeStamp,
    DataType,
    MultiSeries,
    TimeRange,
)
from synnax.util.normalize import normalize


class Channel(Payload):
    """A channel is a logical collection of samples emitted by or representing the
    values of a single source. See
    https://docs.synnaxlabs.com/reference/concepts/channels for an introduction to
    channels and how they work.
    """

    ___frame_client: framer.Client | None = PrivateAttr(None)
    __client: Client | None = PrivateAttr(None)

    def __init__(
        self,
        *,
        name: str,
        data_type: CrudeDataType,
        is_index: bool = False,
        index: Key = 0,
        leaseholder: int = 0,
        key: Key = 0,
        virtual: bool | None = None,
        internal: bool = False,
        expression: str = "",
        operations: list[Operation] | None = None,
        _frame_client: framer.Client | None = None,
        _client: Client | None = None,
    ) -> None:
        """Initializes a new Channel using the given parameters. It's important to note
        that this does not create the Channel in the cluster. To create the channel,
        call client.channels.create(channel).

        :param name: A name for the channel.
        :param data_type: The data type of the samples in the channel. For example, `"float32"`.
        :param is_index: Boolean indicating whether the channel is an index. Index
        channels should have a data type of synnax.TIMESTAMP.
        :param index: The key of the channel that indexes this channel.
        :param leaseholder: The node that holds the lease for this channel. If you don't know
        what this is, leave it at the default value of 0.
        :param virtual: Boolean indicating whether the channel is virtual. Virtual
        channels do not store any data, and are used for streaming purposes only.
        :param expression: An optional Lua expression that defines the channel as a
        calculation of another channel. If this is set, the channel will be
        automatically configured as virtual.
        :param internal: Boolean indicating whether the channel is internal. Internal
        channels are not visible to the user and are used for internal purposes only.
        :param operations: A list of operations to apply to the channel. Operations
        include aggregations like min, max, avg over a time duration or triggered by
        a reset channel.
        :returns: The created channel.
        :param _frame_client: The backing client for reading and writing data to and
        from the channel. This is provided by the Synnax py during calls to
        .channels.create() and .channels.retrieve() and should not be set by the caller.
        """
        if virtual is None:
            virtual = len(expression) > 0
        super().__init__(
            data_type=DataType(data_type),
            name=name,
            leaseholder=leaseholder,
            key=key,
            is_index=is_index,
            index=index,
            internal=internal,
            virtual=virtual,
            expression=expression,
            operations=operations,
        )
        self.___frame_client = _frame_client
        self.__client = _client

    @overload
    def read(
        self,
        start_or_range: TimeRange,
    ) -> MultiSeries: ...

    @overload
    def read(
        self,
        start_or_range: CrudeTimeStamp,
        end: CrudeTimeStamp,
    ) -> MultiSeries: ...

    def read(
        self,
        start_or_range: CrudeTimeStamp | TimeRange,
        end: CrudeTimeStamp | None = None,
    ) -> MultiSeries:
        """Reads telemetry from the channel between the two timestamps.

        :param start_or_range: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A tuple containing a numpy array of the telemetry and a TimeRange
        representing the range of telemetry. The start of the time range represents
        the timestamp of the first sample in the array.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        tr = TimeRange(start_or_range, end)
        return self.__frame_client.read(tr, self.key)

    def write(self, start: CrudeTimeStamp, data: CrudeSeries) -> None:
        """Writes telemetry to the channel starting at the given timestamp.

        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        self.__frame_client.write(start, self.key, data)

    def rename(self, name: str) -> None:
        """Renames the channel.

        :param name: The new name for the channel.
        :returns: None.
        """
        if self.__client is None:
            raise ValidationError("Cannot rename a channel that has not been created.")
        self.__client.rename(self.key, name)

    @property
    def ontology_id(self) -> ID:
        return ontology_id(self.key)

    @property
    def __frame_client(self) -> framer.Client:
        if self.___frame_client is None:
            raise ValidationError(
                "Cannot read from or write to channel that has not been created."
            )
        return self.___frame_client

    def __hash__(self) -> int:
        return hash(self.key)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Channel):
            return NotImplemented
        return self.key == other.key

    def to_payload(self) -> Payload:
        return Payload(
            data_type=self.data_type,
            name=self.name,
            leaseholder=self.leaseholder,
            key=self.key,
            index=self.index,
            is_index=self.is_index,
            virtual=self.virtual,
            internal=self.internal,
            expression=self.expression,
            operations=self.operations,
        )


class Client:
    """The core py class for executing channel operations against a Synnax cluster."""

    _frame_client: framer.Client
    _retriever: Retriever
    _creator: Writer

    def __init__(
        self,
        frame_client: framer.Client,
        retriever: Retriever,
        creator: Writer,
    ):
        self._frame_client = frame_client
        self._retriever = retriever
        self._creator = creator

    def delete(self, channels: Params) -> None:
        """Deletes on or more channels from the cluster"""
        self._creator.delete(channels)

    @overload
    def create(
        self,
        *,
        data_type: CrudeDataType = DataType.UNKNOWN,
        name: str = "",
        index: Key = 0,
        is_index: bool = False,
        leaseholder: int = 0,
        virtual: bool | None = None,
        expression: str = "",
        operations: list[Operation] | None = None,
        retrieve_if_name_exists: bool = False,
    ) -> Channel: ...

    @overload
    def create(
        self, channels: Channel, *, retrieve_if_name_exists: bool = False
    ) -> Channel: ...

    @overload
    def create(
        self, channels: list[Channel], *, retrieve_if_name_exists: bool = False
    ) -> list[Channel]: ...

    def create(
        self,
        channels: Channel | list[Channel] | None = None,
        *,
        data_type: CrudeDataType = DataType.UNKNOWN,
        name: str = "",
        is_index: bool = False,
        index: Key = 0,
        leaseholder: int = 0,
        virtual: bool | None = None,
        expression: str = "",
        operations: list[Operation] | None = None,
        retrieve_if_name_exists: bool = False,
    ) -> Channel | list[Channel]:
        """Creates new channel(s) in the Synnax cluster.

        Overload 1:
        :param data_type: The data type of the samples in the channel. For example, `"float32"`.
        :param name: A name for the channel.
        :param is_index: Boolean indicating whether the channel is an index. Index
        channels should have a data type of synnax.TIMESTAMP.
        :param index: The key of the channel that indexes this channel.
        :param leaseholder: The node that holds the lease for this channel. If you don't know
        what this is, leave it at the default value of 0.
        :param virtual: Boolean indicating whether the channel is virtual. Virtual
        channels do not store any data, and are used for streaming purposes only.
        :param expression: An optional expression that defines the channel as a
        calculation of another channel. If this is set, the channel will be
        automatically configured as virtual.
        set if expression is not an empty string. If expression is not an empty string,
        this should have at least one channel.
        :param retrieve_if_name_exists: Boolean indicating whether to retrieve channels
        with the same name if they already exist in the cluster.
        :returns: The created channel.

        Overload 2:

        :param channels: A single channel to create.
        :param retrieve_if_name_exists: Boolean indicating whether to retrieve channels
        with the same name if they already exist in the cluster.
        :returns: The created channel.

        Overload 3:

        :param channels: A list of channels to create.
        :param retrieve_if_name_exists: Boolean indicating whether to retrieve channels
        with the same name if they already exist in the cluster.
        :returns: The created channels.
        """

        if channels is None:
            if is_index and data_type == DataType.UNKNOWN:
                data_type = DataType.TIMESTAMP
            _channels = [
                Payload(
                    name=name,
                    leaseholder=leaseholder,
                    data_type=DataType(data_type),
                    index=index,
                    is_index=is_index,
                    virtual=virtual if virtual is not None else len(expression) > 0,
                    expression=expression,
                    operations=operations,
                )
            ]
        elif isinstance(channels, Channel):
            _channels = [channels.to_payload()]
        else:
            _channels = [c.to_payload() for c in channels]

        created = list()
        if retrieve_if_name_exists:
            created = self.__sugar(
                self._retriever.retrieve([ch.name for ch in _channels])
            )
            _channels = [
                c for c in _channels if c.name not in [ch.name for ch in created]
            ]

        created.extend(self.__sugar(self._creator.create(_channels)))
        return created if isinstance(channels, list) else created[0]

    @overload
    def retrieve(self, channel: Key | str) -> Channel: ...

    @overload
    def retrieve(
        self,
        channel: list[Key] | tuple[Key] | list[str] | tuple[str],
    ) -> list[Channel]: ...

    def retrieve(self, channel: Params) -> Channel | list[Channel]:
        """Retrieves a channel or set of channels from the cluster.

        Overload 1:

        :param channel: The key or name of the channel to retrieve. If this is
        :returns: The associated channel.
        :raises QueryError: If the channel is not found.

        Overload 2 + 3:
        :param channel: The list of keys or the list of names for the channels to retrieve.
        :returns: The retrieved channels.
        """
        normal = normalize_params(channel)
        res = self._retriever.retrieve(channel)
        sug = self.__sugar(res)
        if not normal.single:
            return sug
        if len(res) == 1:
            return sug[0]
        if len(res) > 1:
            raise _multiple_results_error(channel, res)
        raise NotFoundError(f"Channel matching '{channel}' not found.")

    @overload
    def rename(self, keys: Key, names: str) -> None:
        """Renames a channel in the cluster.

        :param keys: The key of the channel to rename.
        :param names: The new name for the channel.
        :returns: None.
        """
        ...

    @overload
    def rename(
        self, keys: list[Key] | tuple[Key], names: list[str] | tuple[str]
    ) -> None:
        """Renames one or more channels in the cluster.

        :param keys: The keys of the channels to rename.
        :param names: The new names for the channels.
        :returns: None.
        """
        ...

    def rename(
        self,
        keys: Key | list[Key] | tuple[Key],
        names: str | list[str] | tuple[str],
    ) -> None:
        """Renames one or more channels in the cluster.

        :param keys: The keys of the channels to rename.
        :param names: The new names for the channels.
        :returns: None.
        """
        self._creator.rename(normalize(keys), normalize(names))

    def __sugar(self, channels: list[Payload]) -> list[Channel]:
        return [
            Channel(**c.model_dump(), _frame_client=self._frame_client)
            for c in channels
        ]


def _multiple_results_error(
    channel: Params,
    results: list[Payload],
) -> MultipleFoundError:
    msg = f"""

{len(results)} channels matching '{channel}' found. If you'd like to retrieve all
of them, pass in '{channel}' as an array i.e. ['{channel}'] instead of {channel}.

The channels found were:
    """

    # append a max of five results to the error message. If we have more than five,
    # we'll just say "and x more" at the end.
    for i in range(min(5, len(results))):
        msg += f"{str(results[i])}, "

    if len(results) > 5:
        msg += f"and {len(results) - 5} more."

    return MultipleFoundError(msg)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import argparse
import threading
from abc import abstractmethod

import synnax as sy
from examples.simulators.simulator import Simulator


class SimDAQ(Simulator):
    """Base class for thread-based hardware simulators that write to Synnax."""

    description: str = "Run simulator standalone"
    end_cmd_channel: str | None = None

    def __init__(self, client: sy.Synnax, verbose: bool = False):
        super().__init__(verbose=verbose)
        self.client = client
        self._thread: threading.Thread | None = None
        self._end_cmd_thread: threading.Thread | None = None

    def start(self) -> None:
        """Create channels and start simulation loop in background thread."""
        self._create_channels()
        self._running = True

        if self.end_cmd_channel is not None:
            self._end_cmd_thread = threading.Thread(
                target=self._watch_end_cmd, daemon=True
            )
            self._end_cmd_thread.start()

        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._thread.start()

    def _watch_end_cmd(self) -> None:
        """Watch for end command and stop the simulator."""
        with self.client.open_streamer([self.end_cmd_channel]) as streamer:
            while self._running:
                frame = streamer.read(timeout=1)
                if frame is not None:
                    data = frame.get(self.end_cmd_channel)
                    if len(data) > 0:
                        val = data[-1]
                        if hasattr(val, "item"):
                            val = val.item()
                        if val != 0:
                            self.log(f"Received end command on {self.end_cmd_channel}")
                            self._running = False
                            return

    def stop(self, timeout: sy.TimeSpan = 5 * sy.TimeSpan.SECOND) -> None:
        """Stop simulation and wait for threads to finish."""
        self._running = False
        timeout_secs = float(timeout / sy.TimeSpan.SECOND)
        if self._end_cmd_thread:
            self._end_cmd_thread.join(timeout=timeout_secs)
        if self._thread:
            self._thread.join(timeout=timeout_secs)

    @abstractmethod
    def _create_channels(self) -> None:
        """Create required Synnax channels."""
        pass

    @abstractmethod
    def _run_loop(self) -> None:
        """Main simulation loop. Check self._running to know when to stop."""
        pass

    @classmethod
    def main(cls) -> None:
        """Standalone execution entry point. Call from if __name__ == '__main__'."""
        parser = argparse.ArgumentParser(description=cls.description)
        parser.add_argument("--host", default="", help="Synnax server host")
        parser.add_argument("--port", type=int, default=0, help="Synnax server port")
        parser.add_argument("--username", default="", help="Synnax username")
        parser.add_argument("--password", default="", help="Synnax password")
        args = parser.parse_args()

        # Use saved credentials if no host provided, otherwise use CLI args
        if args.host:
            client = sy.Synnax(
                host=args.host,
                port=args.port,
                username=args.username,
                password=args.password,
            )
        else:
            client = sy.Synnax()

        print(f"Starting {cls.__name__} (Ctrl+C to stop)...")
        sim = cls(client, verbose=True)
        sim.start()

        try:
            while sim._running:
                sy.sleep(1)
        except KeyboardInterrupt:
            print("\nStopping simulator...")
        finally:
            sim.stop()
            print("Done.")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import numpy as np
import pandas as pd
import pytest

import synnax as sy
from synnax.util.random import random_name
from tests.telem import seconds_linspace


@pytest.mark.framer
@pytest.mark.delete
class TestDeleter:
    def test_basic_delete(
        self, indexed_pair: tuple[sy.Channel, sy.Channel], client: sy.Synnax
    ):
        idx_ch, _ = indexed_pair
        with client.open_writer(sy.TimeSpan.SECOND * 1, idx_ch) as w:
            data = seconds_linspace(1, 50)
            w.write(pd.DataFrame({idx_ch.key: data}))
            w.commit()

        client.delete(
            [idx_ch.key],
            sy.TimeRange(0, sy.TimeStamp(25 * sy.TimeSpan.SECOND)),
        )

        data = idx_ch.read(sy.TimeRange.MAX)
        assert data.to_numpy().size == 26
        assert data.time_range == sy.TimeRange(
            sy.TimeStamp(25 * sy.TimeSpan.SECOND),
            sy.TimeStamp(50 * sy.TimeSpan.SECOND) + 1,
        )

    def test_delete_by_name(self, indexed_pair: list[sy.Channel], client: sy.Synnax):
        idx_ch, _ = indexed_pair
        with client.open_writer(sy.TimeSpan.SECOND * 1, idx_ch) as w:
            data = seconds_linspace(1, 50)
            w.write(pd.DataFrame({idx_ch.key: data}))
            w.commit()

        client.delete(
            [idx_ch.name],
            sy.TimeRange(0, sy.TimeStamp(25 * sy.TimeSpan.SECOND)),
        )

        data = idx_ch.read(sy.TimeRange.MAX)
        assert data.to_numpy().size == 26
        assert data.time_range == sy.TimeRange(
            sy.TimeStamp(25 * sy.TimeSpan.SECOND),
            sy.TimeStamp(50 * sy.TimeSpan.SECOND) + 1,
        )

    def test_delete_channel_not_found_name(
        self, indexed_pair: tuple[sy.Channel, sy.Channel], client: sy.Synnax
    ):
        idx_ch, _ = indexed_pair
        client.write(0, idx_ch.key, seconds_linspace(1, 50))
        with pytest.raises(sy.NotFoundError):
            client.delete([idx_ch.name, "kaka"], sy.TimeRange.MAX)

        data = idx_ch.read(sy.TimeRange.MAX)
        assert data.size == 50 * 8

    def test_delete_channel_not_found_key(
        self, indexed_pair: tuple[sy.Channel, sy.Channel], client: sy.Synnax
    ):
        idx_ch, _ = indexed_pair
        client.write(0, idx_ch.key, seconds_linspace(1, 50))
        with pytest.raises(sy.NotFoundError):
            client.delete([idx_ch.key, 1234], sy.TimeRange.MAX)

        data = idx_ch.read(sy.TimeRange.MAX)
        assert data.size == 50 * 8

    def test_delete_with_writer(
        self, indexed_pair: list[sy.Channel], client: sy.Synnax
    ):
        idx_ch, _ = indexed_pair
        with client.open_writer(0, idx_ch.key):
            with pytest.raises(sy.UnauthorizedError):
                client.delete(
                    [idx_ch.key],
                    sy.TimeRange(
                        sy.TimeStamp(1 * sy.TimeSpan.SECOND).range(
                            sy.TimeStamp(2 * sy.TimeSpan.SECOND)
                        )
                    ),
                )

    def test_delete_index_alone(self, client: sy.Synnax):
        ch1 = client.channels.create(
            sy.Channel(
                name=random_name(), data_type=sy.DataType.TIMESTAMP, is_index=True
            )
        )

        ch2 = sy.Channel(
            name=random_name(),
            data_type=sy.DataType.FLOAT32,
            index=ch1.key,
        )

        ch2 = client.channels.create(ch2)
        timestamps = [
            sy.TimeStamp(0),
            sy.TimeStamp(1 * sy.TimeSpan.SECOND),
            sy.TimeStamp(2 * sy.TimeSpan.SECOND),
            sy.TimeStamp(3 * sy.TimeSpan.SECOND),
            sy.TimeStamp(4 * sy.TimeSpan.SECOND),
        ]
        ch1.write(0, np.array(timestamps))
        ch2.write(0, np.array([0, 1, 2, 3, 4], dtype=np.float32))

        with pytest.raises(Exception):
            client.delete(
                [ch1.key],
                sy.TimeRange(
                    sy.TimeStamp(1 * sy.TimeSpan.SECOND).range(
                        sy.TimeStamp(2 * sy.TimeSpan.SECOND)
                    )
                ),
            )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import time

import numpy as np
import pandas as pd
import pytest

import synnax as sy
from tests.channel import assert_eventually_channels_are_found


@pytest.mark.framer
@pytest.mark.streamer
class TestStreamer:
    def test_basic_stream_virtual(self, virtual_channel: sy.Channel, client: sy.Synnax):
        """Should correctly stream data for a virtual channel"""
        with client.open_streamer(virtual_channel.key) as s:
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                for i in range(10):
                    data = np.random.rand(10).astype(np.float64)
                    w.write(pd.DataFrame({virtual_channel.key: data}))
                    frame = s.read(timeout=1)
                    assert np.array_equal(frame[virtual_channel.key], data)

    def test_basic_stream_indexed_pair(
        self,
        indexed_pair: list[sy.Channel],
        client: sy.Synnax,
    ):
        idx, data = indexed_pair
        """Should correctly stream data from a virtual/index channel pair"""
        with client.open_streamer(channels=[idx.name, data.name]) as s:
            with client.open_writer(sy.TimeStamp.now(), channels=indexed_pair) as w:
                for i in range(10):
                    ts = sy.TimeStamp.now()
                    value = np.random.rand(1)
                    w.write(pd.DataFrame({idx.name: ts, data.name: value}))
                    frame = s.read(timeout=1)
                    assert frame[idx.name][0] == ts
                    assert frame[data.name][0] == value

    def test_basic_stream_multiple_writers(
        self,
        indexed_pair: list[sy.Channel],
        virtual_channel: sy.Channel,
        client: sy.Synnax,
    ):
        idx, data = indexed_pair
        start = sy.TimeStamp.now()
        with client.open_streamer(
            channels=[idx.name, data.name, virtual_channel.name]
        ) as s:
            with client.open_writer(start, channels=indexed_pair) as idx_writer:
                with client.open_writer(
                    start,
                    channels=virtual_channel,
                ) as virtual_writer:
                    for i in range(10):
                        ts = sy.TimeStamp.now()
                        value = np.random.rand(1)
                        v_value = np.random.rand(1)
                        idx_writer.write(pd.DataFrame({idx.name: ts, data.name: value}))
                        virtual_writer.write(
                            pd.DataFrame({virtual_channel.name: v_value})
                        )
                        for _ in range(2):
                            frame = s.read(timeout=1)
                            if len(frame.channels) == 1:
                                assert frame[virtual_channel.name] == v_value
                            else:
                                assert frame[idx.name][0] == ts
                                assert frame[data.name][0] == value

    def test_open_streamer_no_channels(self, client: sy.Synnax):
        """Should not throw an exception when a streamer is opened with no channels"""
        with client.open_streamer([]):
            pass

    def test_open_streamer_channel_not_found(self, client: sy.Synnax):
        """Should throw an exception when a streamer is opened with an unknown channel"""
        with pytest.raises(sy.NotFoundError):
            with client.open_streamer([123]):
                pass

    def test_open_streamer_channel_key_zero(self, client: sy.Synnax):
        """Should throw an exception when a streamer is opened with a channel key of
        zero"""
        with pytest.raises(sy.NotFoundError):
            with client.open_streamer([0, 0, 0]):
                pass

    def test_update_channels(self, virtual_channel: sy.Channel, client: sy.Synnax):
        """Should update the list of channels to stream"""
        with client.open_streamer([]) as s:
            s.update_channels([virtual_channel.key])
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                data = np.random.rand(1).astype(np.float64)
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = s.read(timeout=1)
                assert np.array_equal(frame[virtual_channel.key], data)

    def test_update_channels_rapid(
        self,
        indexed_pair: list[sy.Channel],
        virtual_channel: sy.Channel,
        client: sy.Synnax,
    ):
        idx, data_ch = indexed_pair
        start = sy.TimeStamp.now()
        channel_names = [idx.name, data_ch.name, virtual_channel.name]
        curr_channels = [*channel_names]
        with client.open_streamer(channels=channel_names) as s:
            with client.open_writer(start, channels=channel_names) as writer:
                for i in range(300):
                    ts = sy.TimeStamp.now()
                    value = np.random.rand(1)
                    data = {
                        idx.name: ts,
                        data_ch.name: value,
                        virtual_channel.name: value,
                    }
                    writer.write(data)
                    if i % 50 == 0:
                        c = int((i % 150) / 50) + 1
                        curr_channels = channel_names[:c]
                        s.update_channels(curr_channels)
                    frm = s.read(timeout=1)
                    for channel in curr_channels:
                        d = frm.get(channel, None)
                        if len(d) > 0:
                            assert len(d) == 1
                            assert d[0] == data[channel]

    def test_timeout_seconds(self, client: sy.Synnax):
        """Should return None after the specified timeout is exceeded"""
        with client.open_streamer([]) as s:
            start = sy.TimeStamp.now()
            f = s.read(timeout=0.1)
            assert f is None
            assert abs(sy.TimeSpan.since(start).seconds - 0.1) < 0.05

    def test_timeout_timespan(self, client: sy.Synnax):
        """Should return None after the specified timeout is exceeded"""
        with client.open_streamer([]) as s:
            start = sy.TimeStamp.now()
            f = s.read(timeout=100 * sy.TimeSpan.MILLISECOND)
            assert f is None
            assert abs(sy.TimeSpan.since(start).seconds - 0.1) < 0.05

    def test_downsample(self, virtual_channel: sy.Channel, client: sy.Synnax):
        """Should correctly stream data for a channel"""
        with client.open_streamer(virtual_channel.key, 1) as s:
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                data = np.random.rand(10).astype(np.float64)
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = s.read(timeout=1)
                assert np.array_equal(frame[virtual_channel.key], data)
        with client.open_streamer(virtual_channel.key, 2) as s:
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                expect = [1.0, 3.0, 5.0, 7.0, 9.0]
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = s.read(timeout=1)
                assert np.array_equal(frame[virtual_channel.key], expect)
        with client.open_streamer(virtual_channel.key, 10) as s:
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                expect = [1.0]
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = s.read(timeout=1)
                assert np.array_equal(frame[virtual_channel.key], expect)
        with client.open_streamer(virtual_channel.key, 20) as s:
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                expect = [1.0]
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = s.read(timeout=1)
                assert np.array_equal(frame[virtual_channel.key], expect)

    def test_downsample_negative(self, virtual_channel: sy.Channel, client: sy.Synnax):
        with pytest.raises(sy.ValidationError):
            with client.open_streamer(virtual_channel.key, -1) as s:
                ...

    @pytest.mark.multi_node
    def test_multi_node_stream_case_1(self):
        client = sy.Synnax(cache_channels=False)
        idx = client.channels.create(
            name="idx", data_type=sy.DataType.TIMESTAMP, is_index=True, leaseholder=2
        )
        data = client.channels.create(
            name="data", data_type=sy.DataType.FLOAT64, index=idx.key, leaseholder=2
        )
        assert_eventually_channels_are_found(client, [idx.key, data.key])
        with client.open_streamer(data.key) as s:
            with client.open_writer(sy.TimeStamp.now(), [idx.key, data.key]) as w:
                w.write({idx.key: [sy.TimeStamp.now()], data.key: [1]})
                f = s.read(timeout=1)
                assert f is not None
                assert f[data.key][0] == 1

    @pytest.mark.multi_node
    def test_multi_node_stream_case_2(self):
        node_1_client = sy.Synnax(
            host="localhost",
            port=9090,
            username="synnax",
            password="seldon",
            secure=False,
            cache_channels=False,
        )
        node_2_client = sy.Synnax(
            host="localhost",
            port=9091,
            username="synnax",
            password="seldon",
            secure=False,
            cache_channels=False,
        )
        idx = node_2_client.channels.create(
            name="idx", data_type=sy.DataType.TIMESTAMP, is_index=True, leaseholder=2
        )
        data = node_2_client.channels.create(
            name="data", data_type=sy.DataType.FLOAT64, index=idx.key, leaseholder=2
        )
        assert_eventually_channels_are_found(node_1_client, [idx.key, data.key])
        assert_eventually_channels_are_found(node_2_client, [idx.key, data.key])
        with node_1_client.open_streamer(data.key) as s:
            with node_2_client.open_writer(
                sy.TimeStamp.now(), [idx.key, data.key]
            ) as w:
                w.write({idx.key: [sy.TimeStamp.now()], data.key: [1]})
                f = s.read(timeout=1)
                assert f is not None
                assert f[data.key][0] == 1


@pytest.mark.framer
class TestAsyncStreamer:
    @pytest.mark.asyncio
    async def test_basic_stream(self, virtual_channel: sy.Channel, client: sy.Synnax):
        with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
            async with await client.open_async_streamer(virtual_channel.key, 1) as s:
                time.sleep(0.1)
                data = np.random.rand(10).astype(np.float64)
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = await s.read()
                assert np.array_equal(frame[virtual_channel.key], data)
        with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
            async with await client.open_async_streamer(virtual_channel.key, 2) as s:
                time.sleep(0.1)
                data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                expect = [1.0, 3.0, 5.0, 7.0, 9.0]
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = await s.read()
                assert np.array_equal(frame[virtual_channel.key], expect)
        with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
            async with await client.open_async_streamer(virtual_channel.key, 10) as s:
                time.sleep(0.1)
                data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                expect = [1.0]
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = await s.read()
                assert np.array_equal(frame[virtual_channel.key], expect)
        with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
            async with await client.open_async_streamer(virtual_channel.key, 20) as s:
                time.sleep(0.1)
                data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                expect = [1.0]
                w.write(pd.DataFrame({virtual_channel.key: data}))
                frame = await s.read()
                assert np.array_equal(frame[virtual_channel.key], expect)

    @pytest.mark.asyncio
    async def test_downsample_negative(
        self, virtual_channel: sy.Channel, client: sy.Synnax
    ):
        with pytest.raises(sy.ValidationError):
            with client.open_writer(sy.TimeStamp.now(), virtual_channel.key) as w:
                async with await client.open_async_streamer(
                    virtual_channel.key, -1
                ) as s:
                    time.sleep(0.1)
                    data = [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                    w.write(pd.DataFrame({virtual_channel.key: data}))
                    frame = await s.read()
                    assert np.array_equal(frame[virtual_channel.key], data)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import UUID

from pydantic import BaseModel

from synnax import task

TASK_TYPE = "arc"


class TaskConfig(BaseModel):
    """Configuration for an Arc task."""

    arc_key: str
    """The key of the Arc program to execute (UUID as string)."""
    auto_start: bool = False
    """Whether to start the task automatically when created."""


class Task(task.StarterStopperMixin, task.JSONConfigMixin, task.Protocol):
    """A task for executing Arc programs on a Synnax cluster."""

    TYPE = TASK_TYPE
    config: TaskConfig
    _internal: task.Task

    def __init__(
        self,
        internal: task.Task | None = None,
        *,
        name: str = "Arc Task",
        arc_key: UUID | str | None = None,
        auto_start: bool = False,
    ):
        """Initialize an Arc task.

        :param internal: Internal task object (used when loading from server).
        :param name: Human-readable name for the task.
        :param arc_key: The key of the Arc program to execute.
        :param auto_start: Whether to start the task automatically.
        """
        if internal is not None:
            self._internal = internal
            self.config = TaskConfig.model_validate(internal.config)
            return
        if arc_key is None:
            raise ValueError("arc_key is required when creating a new ArcTask")
        self._internal = task.Task(name=name, type=self.TYPE)
        self.config = TaskConfig(arc_key=str(arc_key), auto_start=auto_start)


from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "ArcTask": "Task",
    "ArcTaskConfig": "TaskConfig",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy


def seconds_linspace(start: int, count: int) -> list[sy.TimeSpan]:
    """Generates a list of TimeSpan values from start to start + count"""
    return [start * sy.TimeSpan.SECOND + i * sy.TimeSpan.SECOND for i in range(count)]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to read data from a LabJack device.

Before running this example:
1. Connect your LabJack device (T4, T7, T7-Pro, T8, or Digit) to your computer

2. Register the device in Synnax:
   uv run python examples/labjack/connect_device.py

3. Connect analog sensors to AIN0 and AIN1 (or update channel configuration below)

4. Run this script:
   uv run python examples/labjack/read_task.py

This example reads from two analog input channels at 100 Hz.
"""

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Retrieve the LabJack device from Synnax
# Update this with the name you gave the device in the Synnax Console
dev = client.devices.retrieve(name="My LabJack T7")

# Create an index channel that will be used to store the timestamps for the data.
labjack_time = client.channels.create(
    name="labjack_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create Synnax channels to store the analog input data.
ain0 = client.channels.create(
    name="labjack_ain0",
    index=labjack_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)

ain1 = client.channels.create(
    name="labjack_ain1",
    index=labjack_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)

# Create the LabJack Read Task
# Reads two analog input channels at 100 Hz
tsk = sy.labjack.ReadTask(
    name="LabJack Py - Read Task",
    device=dev.key,
    sample_rate=sy.Rate.HZ * 100,  # Sample at 100 Hz
    stream_rate=sy.Rate.HZ * 25,  # Stream at 25 Hz (4 samples per stream)
    data_saving=True,
    channels=[
        # Analog input channel AIN0 with ±10V range
        sy.labjack.AIChan(
            port="AIN0",
            channel=ain0.key,
            range=10.0,
            pos_chan=0,
            neg_chan=199,  # 199 = single-ended (GND)
        ),
        # Analog input channel AIN1 with ±10V range
        sy.labjack.AIChan(
            port="AIN1",
            channel=ain1.key,
            range=10.0,
            pos_chan=1,
            neg_chan=199,  # 199 = single-ended (GND)
        ),
    ],
)

# Configure the task with Synnax
try:
    client.tasks.configure(tsk)
    print("✓ Task configured successfully")
except Exception as e:
    print(f"✗ Task configuration failed: {e}")
    exit(1)

print("=" * 70)
print("Starting LabJack Read Task")
print("=" * 70)
print("Reading analog inputs from AIN0 and AIN1...")
print("Running continuously - Press Ctrl+C to stop\n")

print(f"{'Sample':<10} {'Time (s)':<10} {'AIN0 (V)':>12} {'AIN1 (V)':>12}")
print("-" * 70)

# Start the task and read data continuously
try:
    # Hide cursor for clean output
    print("\033[?25l", end="", flush=True)

    with tsk.run():
        with client.open_streamer(["labjack_ain0", "labjack_ain1"]) as streamer:
            sample_count = 0
            start_time = sy.TimeStamp.now()

            while True:
                frame = streamer.read()
                if frame:
                    # Print the latest values from both channels
                    if "labjack_ain0" in frame and len(frame["labjack_ain0"]) > 0:
                        val0 = frame["labjack_ain0"][-1]
                        val1 = frame["labjack_ain1"][-1]
                        elapsed = sy.TimeStamp.now().span(start_time).seconds

                        sample_count += 1
                        print(
                            f"{sample_count:<10} {elapsed:<10.1f} {val0:>12.3f} {val1:>12.3f}",
                            end="\r",
                            flush=True,
                        )

# Output summary
except KeyboardInterrupt:
    print("\n" + "-" * 70)
    print("✓ Read task stopped by user")
    print(f"\nCollected {sample_count} samples")
    print("=" * 70)
finally:
    # Ensure cursor is always shown even if something goes wrong
    print("\033[?25h", end="", flush=True)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This script deletes a Modbus TCP server from Synnax.

If the server is not found, the script exits with an appropriate message.
Otherwise, it prompts for confirmation before deleting.

Before running this example:
1. Start the Synnax Driver (if not already running).

2. Login to Synnax (if not already logged in):
   uv run sy login

3. Run this script:
   uv run python examples/modbus/delete_modbus_server.py

Note: You do NOT need the Modbus TCP server running to delete the server registration.
      This script only removes the server registration from Synnax.

Configuration:
    Modify the constants below to match your Modbus server name.
"""

import synnax as sy

# Configuration
DEVICE_NAME = "Modbus Server"

# Connect to Synnax
client = sy.Synnax()

print("=" * 70)
print("Modbus TCP Server Deletion Script")
print("=" * 70)
print(f"Target Server Name: {DEVICE_NAME}")
print()

# Check if server exists
existing_device = client.devices.retrieve(name=DEVICE_NAME, ignore_not_found=True)

if existing_device is None:
    print("✓ Server not found - nothing to delete.")
    print()
    print("The server is already disconnected from Synnax.")
    print("=" * 70)
    exit(0)

# Server found - show info and prompt for confirmation
print("Server found:")
print(f"- Name: {existing_device.name}")
print(f"- Key: {existing_device.key}")
print(f"- Location: {existing_device.location}")
print()
print("Are you sure you want to delete this server? [Y/N]: ", end="", flush=True)

try:
    response = input().strip().lower()
except (KeyboardInterrupt, EOFError):
    print("\n")
    response = "n"

if response in ("y", "yes"):
    # Delete the server
    print()
    print("Deleting server...")

    try:
        client.devices.delete([existing_device.key])

        print("✓ Server deleted successfully!")
        print()
        print("The Modbus server has been disconnected from Synnax.")
        print("=" * 70)
        exit(0)

    except Exception as e:
        print(f"✗ Failed to delete server: {e}")
        print()
        print("Please try deleting manually in the Synnax Console:")
        print("1. Open Synnax Console")
        print("2. Go to Resources → Devices")
        print(f"3. Find '{DEVICE_NAME}' and delete it")
        print("=" * 70)
        exit(1)

else:
    # User cancelled
    print()
    print("Deletion cancelled - server preserved.")
    print("=" * 70)
    exit(0)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

client = sy.Synnax(
    host="localhost",
    port=9090,
    secure=False,
    username="synnax",
    password="seldon",
)


with client.open_streamer(["sy_task_set"]) as s:
    for frame in s:
        print(frame)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import serial

import synnax as sy

PORT = "/dev/cu.usbmodem21401"
BAUD_RATE = 9600

ser = serial.Serial(PORT, BAUD_RATE)

client = sy.Synnax(
    host="localhost",
    port=9090,
    username="synnax",
    password="seldon",
)

arduino_command = client.channels.create(
    name="arduino_command",
    data_type="uint8",
    virtual=True,
    retrieve_if_name_exists=True,
)

arduino_time = client.channels.create(
    name="arduino_time",
    is_index=True,
    data_type="timestamp",
    retrieve_if_name_exists=True,
)

arduino_state = client.channels.create(
    name="arduino_state",
    index=arduino_time.key,
    data_type="uint8",
    retrieve_if_name_exists=True,
)

arduino_value = client.channels.create(
    name="arduino_value",
    index=arduino_time.key,
    data_type="float32",
    retrieve_if_name_exists=True,
)

with client.open_streamer(["arduino_command"]) as streamer:
    with client.open_writer(
        start=sy.TimeStamp.now(),
        channels=["arduino_time", "arduino_state", "arduino_value"],
    ) as writer:
        while True:
            frame = streamer.read(timeout=0)
            if frame is not None:
                command = str(frame["arduino_command"][0])
                ser.write(command.encode("utf-8"))
            data = ser.readline().decode("utf-8").rstrip()
            if data:
                split = data.split(",")
                writer.write(
                    {
                        "arduino_time": sy.TimeStamp.now(),
                        "arduino_state": int(split[0]),
                        "arduino_value": float(split[1]),
                    }
                )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Any, overload
from uuid import UUID

from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel

from synnax.ontology import ID
from synnax.status.payload import Status
from synnax.util.normalize import normalize
from synnax.util.params import require_named_params


class _SetRequest(BaseModel):
    parent: ID | None = None
    statuses: list[Status[Any]]


class _SetResponse(BaseModel):
    statuses: list[Status[Any]]


class _RetrieveRequest(BaseModel):
    keys: list[str] | None = None
    search_term: str | None = None
    offset: int | None = None
    limit: int | None = None
    include_labels: bool | None = None
    has_labels: list[UUID] | None = None


class _RetrieveResponse(BaseModel):
    statuses: list[Status[Any]] | None = None


class _DeleteRequest(BaseModel):
    keys: list[str]


_SET_ENDPOINT = "/status/set"
_RETRIEVE_ENDPOINT = "/status/retrieve"
_DELETE_ENDPOINT = "/status/delete"


class Client:
    """Client for managing statuses in a Synnax cluster.

    Statuses are standardized messages used to communicate state across the Synnax
    platform. They support different variants (success, info, warning, error, loading,
    disabled) and can be organized hierarchically using the ontology system.
    """

    client: UnaryClient

    def __init__(self, transport: UnaryClient) -> None:
        """Initialize the status client.

        Args:
            transport: The unary transport client to use for communication.
        """
        self.client = transport

    @overload
    def set(
        self,
        status: Status[Any],
        *,
        parent: ID | None = None,
    ) -> Status[Any]: ...

    @overload
    def set(
        self,
        status: list[Status[Any]],
        *,
        parent: ID | None = None,
    ) -> list[Status[Any]]: ...

    def set(
        self,
        status: Status[Any] | list[Status[Any]] | None = None,
        *,
        parent: ID | None = None,
    ) -> Status[Any] | list[Status[Any]]:
        """Create or update a status.

        Args:
            status: A single status or list of statuses to create or update.
            parent: Optional parent ontology ID for organizing the status.

        Returns:
            The created/updated status or list of statuses.

        Examples:
            Create a single status:
                >>> status = client.statuses.set(
                ...     Status(variant="info", message="Task started")
                ... )

            Create multiple statuses:
                >>> statuses = client.statuses.set([
                ...     Status(variant="success", message="Task 1 complete"),
                ...     Status(variant="error", message="Task 2 failed"),
                ... ])

            Create with parent:
                >>> from synnax.status import ontology_id
                >>> parent_id = ontology_id("parent-key")
                >>> status = client.statuses.set(
                ...     Status(variant="info", message="Child status"),
                ...     parent=parent_id
                ... )
        """
        if status is None:
            raise ValueError("status parameter is required")

        single = not isinstance(status, list)
        statuses = [status] if single else status

        res = send_required(
            self.client,
            _SET_ENDPOINT,
            _SetRequest(statuses=statuses, parent=parent),
            _SetResponse,
        ).statuses

        if single:
            return res[0]
        return res

    @overload
    def retrieve(self, *, key: str, include_labels: bool = False) -> Status[Any]: ...

    @overload
    def retrieve(
        self, *, keys: list[str], include_labels: bool = False
    ) -> list[Status[Any]]: ...

    @overload
    def retrieve(
        self,
        *,
        search_term: str | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include_labels: bool = False,
        has_labels: list[UUID] | None = None,
    ) -> list[Status[Any]]: ...

    @require_named_params(example_params=("key", "'status-key-123'"))
    def retrieve(
        self,
        *,
        key: str | None = None,
        keys: list[str] | None = None,
        search_term: str | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include_labels: bool = False,
        has_labels: list[UUID] | None = None,
    ) -> Status[Any] | list[Status[Any]]:
        """Retrieve statuses from the cluster.

        Args:
            key: Retrieve a single status by key.
            keys: Retrieve multiple statuses by keys.
            search_term: Search for statuses by name or message.
            offset: Pagination offset.
            limit: Pagination limit.
            include_labels: Whether to include labels in the response.
            has_labels: Filter statuses that have all of these labels.

        Returns:
            A single status, or a list of statuses.

        Examples:
            Retrieve by key:
                >>> status = client.statuses.retrieve(key="my-status")

            Retrieve multiple:
                >>> statuses = client.statuses.retrieve(
                ...     keys=["status-1", "status-2"]
                ... )

            Search with pagination:
                >>> statuses = client.statuses.retrieve(
                ...     search_term="error",
                ...     limit=10,
                ...     offset=0
                ... )

            Filter by labels:
                >>> from uuid import UUID
                >>> statuses = client.statuses.retrieve(
                ...     has_labels=[UUID("label-uuid")],
                ...     include_labels=True
                ... )
        """
        single = key is not None
        if single and key is not None:
            keys = [key]

        res = send_required(
            self.client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(
                keys=keys,
                search_term=search_term,
                offset=offset,
                limit=limit,
                include_labels=include_labels,
                has_labels=has_labels,
            ),
            _RetrieveResponse,
        ).statuses

        if res is None:
            if single:
                raise ValueError(f"Status with key '{key}' not found")
            return []

        if single:
            if len(res) == 0:
                raise ValueError(f"Status with key '{key}' not found")
            return res[0]
        return res

    def delete(self, keys: str | list[str]) -> None:
        """Delete statuses by their keys.

        This operation is idempotent - deleting a non-existent status will not
        raise an error.

        Args:
            keys: A single key or list of keys to delete.

        Examples:
            Delete a single status:
                >>> client.statuses.delete("my-status")

            Delete multiple statuses:
                >>> client.statuses.delete(["status-1", "status-2", "status-3"])
        """
        send_required(
            self.client,
            _DELETE_ENDPOINT,
            _DeleteRequest(keys=normalize(keys)),
            Empty,
        )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import UUID

from pydantic import BaseModel, ConfigDict

from synnax import ontology
from synnax.color import Color
from synnax.telem import TimeRange

ONTOLOGY_TYPE = ontology.ID(type="range")


def ontology_id(key: UUID) -> ontology.ID:
    """Returns the ontology ID for the Range entity."""
    return ontology.ID(type=ONTOLOGY_TYPE.type, key=str(key))


class Payload(BaseModel):
    """Network transportable payload representing a range."""

    model_config = ConfigDict(validate_assignment=True)

    key: UUID = UUID(int=0)
    name: str = ""
    time_range: TimeRange
    color: Color = Color()


Key = UUID | str
"""The type for the names of a Range. A tuple or list of strings."""
RangeParams = Key | list[Key] | tuple[Key] | str | list[str] | tuple[str]
"""Parameters that can be used to query a range"""

from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "RangePayload": "Payload",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from pydantic import BaseModel


class Options(BaseModel):
    """Options class for the Synnax py."""

    host: str
    port: int
    username: str = ""
    password: str = ""
    secure: bool = False


SynnaxOptions = Options
"""Deprecated: Use Options instead."""



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

"""
This example demonstrates how to configure and start an Analog Write Task on a National
Instruments USB-6289 device.

To run this example, you'll need to have your Synnax cluster properly configured to
detect National Instruments devices: https://docs.synnaxlabs.com/reference/driver/ni/get-started

You'll also need to have either a physical USB-6289 device or create a simulated device
via the NI-MAX software.
"""

# We've logged in via the CLI, so there's no need to provide credentials here.
# See https://docs.synnaxlabs.com/reference/client/quick-start for more information.
client = sy.Synnax()

dev = client.devices.retrieve(model="USB-6289")

# Create virtual command channels that will be used to send commands to the device.
# These are virtual channels that won't store data to disk.
ao_0_cmd = client.channels.create(
    name="ao_0_cmd",
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
    virtual=True,
)
ao_1_cmd = client.channels.create(
    name="ao_1_cmd",
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
    virtual=True,
)

# Create an index channel that will be used to store the timestamps
# for the analog output state data.
ao_state_time = client.channels.create(
    name="ao_state_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create channels that will store the state of the analog outputs
# after they have been commanded.
ao_0_state = client.channels.create(
    name="ao_0_state",
    # Pass in the index key here to associate the channel with the index channel.
    index=ao_state_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)
ao_1_state = client.channels.create(
    name="ao_1_state",
    # Pass in the index key here to associate the channel with the index channel.
    index=ao_state_time.key,
    data_type=sy.DataType.FLOAT32,
    retrieve_if_name_exists=True,
)

# Instantiate the task. A task is a background process that can be used to acquire data
# from, or, in this case, write commands to a device. Tasks are the primary method for
# interacting with Synnax hardware devices.
tsk = sy.ni.AnalogWriteTask(
    # A name to find and monitor the task via the Synnax Console.
    name="Basic Analog Write",
    # The key of the device to execute the task on.
    device=dev.key,
    # The rate at which the task will sample the current state of the analog
    # outputs on the device.
    state_rate=sy.Rate.HZ * 10,
    # Whether to save the states of the analog outputs to disk. If set to False,
    # the data will be streamed into Synnax for real-time consumption but not saved
    # to disk.
    data_saving=True,
    # The mapping of the analog output channels on the device to the Synnax channels.
    channels=[
        sy.ni.AOVoltageChan(
            # The cmd channel will be used to send commands to the device.
            cmd_channel=ao_0_cmd.key,
            # The state channel will be used to store the state of the analog output
            # after it has been commanded.
            state_channel=ao_0_state.key,
            # The port on the device that the analog output is connected to.
            port=0,
            # The minimum and maximum voltage values for the output.
            min_val=-10.0,
            max_val=10.0,
        ),
        sy.ni.AOVoltageChan(
            cmd_channel=ao_1_cmd.key,
            state_channel=ao_1_state.key,
            port=1,
            min_val=-10.0,
            max_val=10.0,
        ),
    ],
)

# Create the task in Synnax and wait for the driver to validate that the configuration
# is correct.
client.tasks.configure(tsk)

# Start the task and write some analog values.
with tsk.run():
    with client.control.acquire(
        name="Analog Control Sequence",
        read=["ao_0_state", "ao_1_state"],
        write=["ao_0_cmd", "ao_1_cmd"],
        write_authorities=50,
    ) as ctrl:
        ctrl["ao_0_cmd"] = 5.0
        ctrl["ao_1_cmd"] = 3.0
        ctrl.wait_until(
            lambda c: abs(c["ao_0_state"] - 5.0) < 0.01
            and abs(c["ao_1_state"] - 3.0) < 0.01,
            timeout=1,
        )
        ctrl.sleep(0.5)

        ctrl["ao_0_cmd"] = -2.5
        ctrl["ao_1_cmd"] = 7.5
        ctrl.wait_until(
            lambda c: abs(c["ao_0_state"] - (-2.5)) < 0.01
            and abs(c["ao_1_state"] - 7.5) < 0.01,
            timeout=1,
        )
        ctrl.sleep(0.5)

        ctrl["ao_0_cmd"] = 0.0
        ctrl["ao_1_cmd"] = 0.0
        ctrl.wait_until(
            lambda c: abs(c["ao_0_state"]) < 0.01 and abs(c["ao_1_state"]) < 0.01,
            timeout=1,
        )

client.tasks.delete(tsk.key)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

client = sy.Synnax()

rack = client.racks.create(name="NI / LabJack Test Rack")

client.devices.create(
    [
        sy.Device(
            key="130227d9-02aa-47e4-b370-0d590add1bc1",
            rack=rack.key,
            name="PXI-6255",
            make="NI",
            model="PXI-6255",
            location="dev1",
        ),
        sy.Device(
            key="labjack-t4",
            rack=rack.key,
            name="LabJack T4",
            make="LabJack",
            model="LJM_dtT4",
            location="dev2",
        ),
        sy.Device(
            key="labjack-t7",
            rack=rack.key,
            name="LabJack T7",
            make="LabJack",
            model="LJM_dtT7",
            location="dev3",
        ),
        sy.Device(
            key="labjack-t8",
            rack=rack.key,
            name="LabJack T8",
            make="LabJack",
            model="LJM_dtT8",
            location="dev4",
        ),
        sy.Device(
            key="a0e37b26-5401-413e-8e65-c7ad9d9afd70",
            rack=rack.key,
            name="USB-6000",
            make="NI",
            model="USB-6000",
            location="dev3",
        ),
    ]
)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
A simple example that creates a large number of channels that are indexed by a single
timestamp channel.
"""

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

CHANNEL_COUNT = 100

# Create an index channel that will be used to store our timestamps.
time_channel = client.channels.create(
    name="create_channels_time",
    data_type=sy.DataType.TIMESTAMP,
    is_index=True,
)

# Create data channels to store our data. Since we did not call client.channels.create
# here, the channels are not actually created in the Synnax cluster yet. We will do that
# in the next step.
data_channels = [
    sy.Channel(
        name=f"create_channels_data_{i}",
        data_type=sy.DataType.FLOAT64,
        index=time_channel.key,
    )
    for i in range(CHANNEL_COUNT)
]

# Notice how we reassign the result of the create call to the data_channels variable.
# This means that all of the channels will have the correct key given to the channel by
# the server.
data_channels = client.channels.create(data_channels)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Sequence, TypeAlias

from pydantic import BaseModel

from synnax import ontology
from synnax.telem import DataType, TimeSpan
from synnax.util.normalize import normalize

Key = int

ONTOLOGY_TYPE = ontology.ID(type="channel")


def ontology_id(key: Key) -> ontology.ID:
    """Returns the ontology ID for the Channel entity."""
    return ontology.ID(type=ONTOLOGY_TYPE.type, key=str(key))


OPERATION_TYPES = Literal["min", "max", "avg", "none"]


class Operation(BaseModel):
    """Represents an operation on a calculated channel."""

    type: OPERATION_TYPES
    reset_channel: Key = 0
    duration: TimeSpan = TimeSpan(0)


class Payload(BaseModel):
    """A payload container that represent the properties of a channel exchanged to and
    from the Synnax server.
    """

    key: Key = 0
    data_type: DataType
    name: str = ""
    leaseholder: int = 0
    is_index: bool = False
    index: Key = 0
    internal: bool = False
    virtual: bool = False
    expression: str | None = ""
    operations: list[Operation] | None = None

    def __str__(self) -> str:
        return f"Channel(name={self.name}, key={self.key})"

    def __hash__(self) -> int:
        return hash(self.key)


@dataclass
class NormalizedKeyResult:
    single: bool
    channels: list[Key] | tuple[Key]


@dataclass
class NormalizedNameResult:
    single: bool
    channels: list[str]


Params: TypeAlias = (
    Key | str | Payload | Sequence[Key] | Sequence[str] | Sequence[Payload]
)


def normalize_params(
    channels: Params,
) -> NormalizedKeyResult | NormalizedNameResult:
    """Determine if a list of keys or names is a single key or name."""
    normalized = normalize(channels)
    if len(normalized) == 0:
        return NormalizedKeyResult(single=False, channels=[])
    single = isinstance(channels, (Key, str))
    first = normalized[0]
    if isinstance(first, str):
        str_list = [s for s in normalized if isinstance(s, str)]
        if len(str_list) != len(normalized):
            raise TypeError(
                "channel params must be all keys or all names, got a mix of both"
            )
        try:
            return NormalizedKeyResult(
                single=single,
                channels=[Key(s) for s in str_list],
            )
        except (ValueError, TypeError):
            return NormalizedNameResult(
                single=single,
                channels=str_list,
            )
    elif isinstance(first, Payload):
        payload_list = [c.key for c in normalized if isinstance(c, Payload)]
        if len(payload_list) != len(normalized):
            raise TypeError(
                "channel params must be all keys or all names, got a mix of both"
            )
        return NormalizedKeyResult(single=single, channels=payload_list)
    key_list = [k for k in normalized if isinstance(k, int)]
    if len(key_list) != len(normalized):
        raise TypeError(
            "channel params must be all keys or all names, got a mix of both"
        )
    return NormalizedKeyResult(single=single, channels=key_list)


def has_params(channels: Params | None) -> bool:
    if channels is None:
        return False
    if isinstance(channels, (Key, str, Payload)):
        return True
    return len(channels) > 0


from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "ChannelKey": "Key",
    "ChannelParams": "Params",
    "ChannelPayload": "Payload",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

# We've logged in via the CLI, so there's no need to provide credentials here.
# See https://docs.synnaxlabs.com/reference/client/quick-start for more information.
client = sy.Synnax()

dev = client.devices.retrieve(model="USB-6289")

# Create an index channel that will be used to store the timestamps
# for the digital read data.
ai_time = client.channels.create(
    name="di_time",
    is_index=True,
    retrieve_if_name_exists=True,
    data_type=sy.DataType.TIMESTAMP,
)
# Create two synnax channels that will be used to store the input data. Notice
# how these channels aren't specifically bound to the device. You'll do that in a
# later step when you create the task.
di_0 = client.channels.create(
    name="di_0",
    # Pass in the index key here to associate the channel with the index channel.
    index=ai_time.key,
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
)
di_1 = client.channels.create(
    name="di_1",
    # Pass in the index key here to associate the channel with the index channel.
    index=ai_time.key,
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
)

# Instantiate the task. A task is a background process that can be used to acquire data
# from, or write commands to a device. Tasks are the primary method for interacting with
# Synnax hardware devices.
tsk = sy.ni.DigitalReadTask(
    # A name to find and monitor the task via the Synnax Console.
    name="Basic Digital Read",
    # The key of the device to execute the task on.
    device=dev.key,
    # The rate at which the task will sample data from the device.
    sample_rate=sy.Rate.HZ * 50,
    # The rate at which data will be streamed from the device into Synnax. Since we're
    # sampling at 50 hz and streaming at 25Hz, we'll get 2 samples at a time.
    stream_rate=sy.Rate.HZ * 25,
    # Whether to save data acquired by the task to disk. If set to True, the data will
    # be streamed into Synnax for real-time consumption but not saved to disk.
    data_saving=True,
    # The list of physical channels we'd like to acquire data from.
    channels=[
        sy.ni.DIChan(channel=di_0.key, port=0, line=0),
        sy.ni.DIChan(channel=di_1.key, port=0, line=1),
    ],
)

# Create the task in Synnax and wait for the driver to validate that the
# configuration is correct.
client.tasks.configure(tsk)

# Stream 100 reads, which will accumulate a total of 200 samples per channel over
# a period of 4 seconds.
total_reads = 100

frame = sy.Frame()

# Start the task under a context manager, which ensures the task gets stopped
# when the block exits.
with tsk.run():
    # Open a streamer on the analog input channels.
    with client.open_streamer(["di_0", "di_1"]) as streamer:
        while total_reads > 0:
            frame.append(streamer.read())
            total_reads -= 1

client.tasks.delete(tsk.key)

# Save the data to a CSV file.
frame.to_df().to_csv("digital_read_result.csv")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to plot data from a Synnax cluster. We will write data to
a Synnax cluster, and then read the data back and plot it.
"""

import matplotlib.pyplot as plt
import numpy as np

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Create an index channel that will be used to store our timestamps.
time_channel = client.channels.create(
    name="plot_time",
    data_type=sy.DataType.TIMESTAMP,
    is_index=True,
)

# Create a data channel that will be used to store our data.
data_channel = client.channels.create(
    name="plot_data",
    data_type=sy.DataType.FLOAT32,
    index=time_channel.key,
)


# We'll start our write at the current time. This timestamp should be the same as or
# before the first timestamp we write.
start = sy.TimeStamp.now()

# We'll end our write 1 minute later
end = start + sy.TimeSpan.MINUTE

# We'll write 1000 samples of data to Synnax.
SAMPLE_COUNT = 1000

# Generate linearly spaced timestamps
timestamps = np.linspace(start, end, SAMPLE_COUNT)

# Generate linearly spaced data with some noise
data = np.linspace(0, 100, SAMPLE_COUNT) + np.random.randn(SAMPLE_COUNT) * 5

# Write the data to the Synnax cluster through the channels.
client.write(
    start,
    {
        time_channel: timestamps,
        data_channel: data,
    },
)

# Create a time range for reading data.
time_range = sy.TimeRange(start, end)

# Read the data back from the Synnax cluster.
data = client.read(time_range, [time_channel.name, data_channel.name])

# For plotting, we want to convert the time to seconds since the start of the range. We
# will convert it to numpy, and then remove the start timestamp value from each sample,
# and then convert from nanoseconds to seconds.
time_data = data[time_channel.name].to_numpy()
time_data = time_data - time_data[0]
time_data = time_data / sy.TimeSpan.SECOND

# Plot the data
plt.plot(time_data, data[data_channel.name])
plt.xlabel(time_channel.name)
plt.ylabel(data_channel.name)
plt.title(f"Plot of {data_channel.name} vs {time_channel.name}")
plt.show()



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import Instrumentation, trace
from freighter import UnaryClient, send_required
from pydantic import BaseModel

import synnax.channel.payload as channel
from synnax.telem import TimeRange


class _Request(BaseModel):
    keys: list[channel.Key] | tuple[channel.Key] | None = None
    names: list[str] | tuple[str] | None = None
    bounds: TimeRange


class _Response(BaseModel): ...


class Deleter:
    """
    Deleter is used to delete a time range of telemetry from the data engine.
    """

    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(self, client: UnaryClient, instrumentation: Instrumentation) -> None:
        self._client = client
        self.instrumentation = instrumentation

    @trace("debug")
    def delete(
        self,
        channels: channel.Params,
        tr: TimeRange,
    ) -> None:
        normal = channel.normalize_params(channels)
        req = _Request(bounds=tr)
        if isinstance(normal, channel.NormalizedKeyResult):
            req.keys = normal.channels
        else:
            req.names = normal.channels
        send_required(self._client, "/frame/delete", req, _Response)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy

# We've logged in via the CLI, so there's no need to provide credentials here.
# See https://docs.synnaxlabs.com/reference/client/quick-start for more information.
client = sy.Synnax()

# Retrieve the USB-6289 device from Synnax.
dev = client.devices.retrieve(model="USB-6289")

# Create a channel that will be used to send commands to the device. We're using
# a virtual channel here that won't store any data to disk. Don't worry, we're
# still going to get data on the state of the digital output.
do_1_cmd = client.channels.create(
    name="do_1_cmd",
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
    virtual=True,
)

# Create a channel that will store timestamps for the state of the digital output.
do_state_time = client.channels.create(
    name="do_state_time",
    is_index=True,
    data_type=sy.DataType.TIMESTAMP,
    retrieve_if_name_exists=True,
)

# Create a channel that will store the state of the digital output.
do_1_state = client.channels.create(
    name="do_1_state",
    # Pass in the index key here to associate the channel with the index channel.
    index=do_state_time.key,
    data_type=sy.DataType.UINT8,
    retrieve_if_name_exists=True,
)

# Instantiate the task. A task is a background process that can be used to acquire data
# from, or, in this case, write commands to a device. Tasks are the primary method for
# interacting with Synnax hardware devices.
tsk = sy.ni.DigitalWriteTask(
    # A name to find and monitor the task via the Synnax Console.
    name="Basic Digital Write",
    # The key of the device to execute the task on.
    device=dev.key,
    # The rate at which the task will sample the current state of the digital
    # outputs on the device.
    state_rate=sy.Rate.HZ * 2000,
    # Whether to save the states of the digital outputs to disk. If set to False,
    # the data will be streamed into Synnax for real-time consumption but not saved
    # to disk.
    data_saving=True,
    # The mapping of the digital output channels on the device to the Synnax channels.
    channels=[
        sy.ni.DOChan(
            # The cmd channel will be used to send commands to the device.
            cmd_channel=do_1_cmd.key,
            # The state channel will be used to store the state of the digital output
            # after it has been commanded.
            state_channel=do_1_state.key,
            # The port and line on the device that the digital output is connected to.
            port=0,
            line=0,
        ),
    ],
)

# Create the task in Synnax and wait for the driver to validate that the configuration
# is correct.
client.tasks.configure(tsk)

durations = list()

# Start the task.
with tsk.run():
    with client.control.acquire(
        name="Control Sequence",
        read=["do_1_state"],
        write=["do_1_cmd"],
        write_authorities=50,
    ) as ctrl:
        for i in range(2500):
            ctrl["do_1_cmd"] = 1
            start = sy.TimeStamp.now()
            ctrl.wait_until(lambda c: c["do_1_state"] == 1)
            durations.append(sy.TimeSpan.since(start))
            ctrl["do_1_cmd"] = 0
            ctrl.wait_until(lambda c: c["do_1_state"] == 0)

client.tasks.delete(tsk.key)

import matplotlib.pyplot as plt
import numpy as np

amortized = np.array([dur.milliseconds for dur in durations]) / 2
# remove the outliers
amortized = amortized[amortized < 2]

# Get the P99 latency
p99 = np.percentile(amortized, 99)
print(f"P99 latency: {p99} ms")
p95 = np.percentile(amortized, 95)
print(f"P95 latency: {p95} ms")
p50 = np.percentile(amortized, 50)
print(f"P50 latency: {p50} ms")

# make the line skinnier and the plot wider
plt.plot(amortized, linewidth=0.5)
# plot the average for every 50 requests in red
plt.plot(np.convolve(amortized, np.ones(50) / 50, mode="valid"), color="red")


plt.xlabel("Request #")
plt.ylabel("Latency (ms)")
plt.savefig("latency.png")

jitter = np.std(amortized)

print(f"Jitter: {jitter} ms")



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.task.client import (
    BaseConfig,
    BaseReadConfig,
    BaseWriteConfig,
    Client,
    JSONConfigMixin,
    Protocol,
    StarterStopperMixin,
    Task,
)
from synnax.task.payload import Payload, Status, StatusDetails
from synnax.util.deprecation import deprecated_getattr

_DEPRECATED = {
    "TaskPayload": "Payload",
    "TaskStatus": "Status",
    "TaskStatusDetails": "StatusDetails",
    "BaseTaskConfig": "BaseConfig",
    "BaseReadTaskConfig": "BaseReadConfig",
    "BaseWriteTaskConfig": "BaseWriteConfig",
    "TaskProtocol": "Protocol",
}

__getattr__ = deprecated_getattr(__name__, _DEPRECATED, globals())

__all__ = [
    "Client",
    "Task",
    "Payload",
    "Status",
    "StatusDetails",
    "BaseConfig",
    "BaseReadConfig",
    "BaseWriteConfig",
    "JSONConfigMixin",
    "StarterStopperMixin",
    "Protocol",
]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from collections.abc import Callable
from functools import wraps

from synnax import channel, framer
from synnax.state import LatestState, State

_InternalHandler = Callable[[State], Callable[[LatestState], None] | None]


class Registry:
    _handlers: list[_InternalHandler]
    _channels: list[channel.Key]
    _frame_client: framer.Client
    _channel_retriever: channel.Retriever

    def __init__(self, frame_client: framer.Client, channels: channel.Retriever):
        self._handlers = list()
        self._channels = list()
        self._frame_client = frame_client
        self._channel_retriever = channels

    def on(
        self,
        channels: channel.Params,
        filter_f: Callable[[LatestState], bool],
    ) -> Callable[[Callable[[LatestState], None]], Callable[[LatestState], None]]:
        normal = channel.normalize_params(channels)
        if isinstance(normal, channel.NormalizedKeyResult):
            self._channels.extend(normal.channels)
        else:
            resolved = self._channel_retriever.retrieve(channels)
            self._channels.extend(ch.key for ch in resolved)

        def decorator(
            f: Callable[[LatestState], None],
        ) -> Callable[[LatestState], None]:
            @wraps(f)
            def wrapper(state: State) -> Callable[[LatestState], None] | None:
                if filter_f(LatestState(state)):
                    return f
                return None

            self._handlers.append(wrapper)
            return f

        return decorator

    async def process(self) -> None:
        await Scheduler(
            channels=self._channels,
            handlers=self._handlers,
            frame_client=self._frame_client,
            channel_retriever=self._channel_retriever,
        ).start()


class Scheduler:
    _streamer: framer.AsyncStreamer | None = None
    _handlers: list[_InternalHandler]
    _channels: list[channel.Key]
    _state: State
    _frame_client: framer.Client
    _channel_retriever: channel.Retriever

    def __init__(
        self,
        channels: list[channel.Key],
        handlers: list[_InternalHandler],
        frame_client: framer.Client,
        channel_retriever: channel.Retriever,
    ):
        self._frame_client = frame_client
        self._channels = channels
        self._handlers = handlers
        self._channel_retriever = channel_retriever
        self._state = State(channel_retriever)

    async def start(self) -> None:
        self._streamer = await self._frame_client.open_async_streamer(self._channels)
        async for frame in self._streamer:
            self._state.update(frame)
            for handler in self._handlers:
                res = handler(self._state)
                if res is not None:
                    res(LatestState(self._state))

    def stop(self) -> None: ...



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from datetime import datetime

import numpy as np

import synnax as sy

TEST_COUNT = 30

client = sy.Synnax()

test_time_channel_1 = client.channels.create(
    name="daq_4189_time",
    is_index=True,
    retrieve_if_name_exists=True,
)

test_time_channel_2 = client.channels.create(
    name="daq_9000_time",
    is_index=True,
    retrieve_if_name_exists=True,
)

test_channel_information = [
    {
        "name": "daq_4189_pressure",
        "count": 10,
        "data_type": sy.DataType.FLOAT32,
        "time_channel": test_time_channel_1.key,
    },
    {
        "name": "daq_4189_temperature",
        "count": 3,
        "data_type": sy.DataType.FLOAT32,
        "time_channel": test_time_channel_1.key,
    },
    {
        "name": "daq_9000_output",
        "count": 5,
        "data_type": sy.DataType.UINT8,
        "time_channel": test_time_channel_2.key,
    },
]

channels_to_create: list[sy.Channel] = []

for info in test_channel_information:
    for i in range(info["count"]):
        channels_to_create.append(
            sy.Channel(
                name=f"{info['name']}_{i}",
                data_type=info["data_type"],
                index=info["time_channel"],
            )
        )

client.channels.create(channels_to_create, retrieve_if_name_exists=True)

all_channels = [test_time_channel_1, test_time_channel_2] + channels_to_create

for i in range(1, TEST_COUNT + 1):
    test_tr = sy.TimeRange(
        start=datetime(2012, 1, i, 13, 0, 0),  # January i, 2012 1:00 PM
        end=datetime(2012, 1, i, 14, 45, 0),  # January i, 2012 3:00 PM
    )
    parent_tr = client.ranges.create(
        name=f"Test January {i}",
        time_range=test_tr,
        color="#E0E0E0",
        retrieve_if_name_exists=True,
    )
    child_tr_1 = parent_tr.create_sub_range(
        name="Pre-test",
        time_range=sy.TimeRange(
            start=datetime(2012, 1, i, 13, 0, 0),
            end=datetime(2012, 1, i, 13, 30, 0),
        ),
        color="#00FF00",
    )
    child_tr_2 = parent_tr.create_sub_range(
        name="Test Run",
        time_range=sy.TimeRange(
            start=datetime(2012, 1, i, 13, 30, 0),
            end=datetime(2012, 1, i, 14, 45, 0),
        ),
        color="#0000FF",
    )
    child_tr_3 = parent_tr.create_sub_range(
        name="Post-test",
        time_range=sy.TimeRange(
            start=datetime(2012, 1, i, 14, 45, 0),
            end=datetime(2012, 1, i, 15, 0, 0),
        ),
        color="#FF00FF",
    )

    # Generate data for each channel during the test run
    daq_4189_series: list[tuple[str, sy.Series]] = []
    daq_9000_series: list[tuple[str, sy.Series]] = []
    for channel in all_channels:
        if "daq_4189" in channel.name:
            if "time" in channel.name:
                # Generate timestamps at 100 Hz for the test run
                samples = int(sy.TimeSpan(test_tr.end - test_tr.start).seconds * 100)
                timestamps = np.linspace(
                    test_tr.start, test_tr.end, samples, dtype=np.int64
                )
                daq_4189_series.append(
                    (
                        channel.name,
                        sy.Series(
                            data=timestamps,
                            time_range=test_tr,
                            data_type=sy.DataType.TIMESTAMP,
                        ),
                    )
                )
                continue

            # Generate 100 Hz sinusoidal data
            samples = int(sy.TimeSpan(test_tr.end - test_tr.start).seconds * 100)
            t = np.linspace(
                0, sy.TimeSpan(test_tr.end - test_tr.start).seconds, samples
            )
            freq = np.random.uniform(0.1, 2.0)  # Random frequency between 0.1 and 2 Hz
            amp = np.random.uniform(0.5, 2.0)  # Random amplitude between 0.5 and 2
            offset = i + np.random.uniform(
                -0.5, 0.5
            )  # Center around i with some variation
            data = amp * np.sin(2 * np.pi * freq * t) + offset
            daq_4189_series.append(
                (
                    channel.name,
                    sy.Series(
                        data=data, time_range=test_tr, data_type=channel.data_type
                    ),
                )
            )

        elif "daq_9000" in channel.name:
            if "time" in channel.name:
                # Generate timestamps at 1 Hz for the test run
                samples = int(sy.TimeSpan(test_tr.end - test_tr.start).seconds)
                timestamps = np.linspace(
                    test_tr.start, test_tr.end, samples, dtype=np.int64
                )
                daq_9000_series.append(
                    (
                        channel.name,
                        sy.Series(
                            data=timestamps,
                            time_range=test_tr,
                            data_type=sy.DataType.TIMESTAMP,
                        ),
                    )
                )
                continue
            # Generate 1 Hz binary data
            samples = int(sy.TimeSpan(test_tr.end - test_tr.start).seconds)
            data = np.random.choice([0, 1], size=samples)

            daq_9000_series.append(
                (
                    channel.name,
                    sy.Series(
                        data=data, time_range=test_tr, data_type=channel.data_type
                    ),
                )
            )
    daq_4189_frame = sy.Frame({key: series for key, series in daq_4189_series})
    daq_9000_frame = sy.Frame({key: series for key, series in daq_9000_series})
    # Write the data to the channels
    client.write(test_tr.start, daq_4189_frame)
    client.write(test_tr.start, daq_9000_frame)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from __future__ import annotations

from uuid import UUID

from pydantic import BaseModel

from synnax import ontology

ONTOLOGY_TYPE = ontology.ID(type="role")


class Role(BaseModel):
    key: UUID | None = None
    name: str
    description: str = ""
    internal: bool = False


def ontology_id(key: UUID | None = None) -> ontology.ID:
    return ontology.ID(type="role", key=key if key is None else str(key))



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import Instrumentation, trace
from freighter import Empty, UnaryClient, send_required
from pydantic import BaseModel

from synnax.channel.payload import (
    Key,
    NormalizedNameResult,
    Params,
    Payload,
    normalize_params,
)
from synnax.channel.retrieve import CacheRetriever


class _CreateRequest(BaseModel):
    channels: list[Payload]


_Response = _CreateRequest


class _DeleteRequest(BaseModel):
    keys: list[Key] | tuple[Key] | None = None
    names: list[str] | tuple[str] | None = None


class _RenameRequest(BaseModel):
    keys: list[Key] | tuple[Key]
    names: list[str] | tuple[str]


class Writer:
    _client: UnaryClient
    _cache: CacheRetriever | None
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation,
        cache: CacheRetriever | None,
    ):
        self._client = client
        self.instrumentation = instrumentation
        self._cache = cache

    @trace("debug")
    def create(
        self,
        channels: list[Payload],
    ) -> list[Payload]:
        req = _CreateRequest(channels=channels)
        res = send_required(self._client, "/channel/create", req, _Response)
        if self._cache is not None:
            self._cache.set(res.channels)
        return res.channels

    @trace("debug")
    def delete(self, channels: Params) -> None:
        normal = normalize_params(channels)
        if isinstance(normal, NormalizedNameResult):
            req = _DeleteRequest(names=normal.channels)
        else:
            req = _DeleteRequest(keys=normal.channels)
        send_required(self._client, "/channel/delete", req, Empty)
        if self._cache is not None:
            self._cache.delete(normal.channels)

    @trace("debug")
    def rename(
        self, keys: list[Key] | tuple[Key], names: list[str] | tuple[str]
    ) -> None:
        req = _RenameRequest(keys=keys, names=names)
        send_required(self._client, "/channel/rename", req, Empty)
        if self._cache is not None:
            self._cache.rename(list(keys), list(names))



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example simulates a basic control sequence where a tank with pressure represented
by press_pt is pressurized in 20 psi increments. The tank is then vented after holding
down to 5 psi after a period of time.

To run this example:

Terminal 1 - Start the simulator:
    cd client/py
    uv run python -m examples.simulators.press

Terminal 2 - Run the control sequence:
    cd client/py
    uv run python -m examples.control.press.control_sequence

This script can be visualized in the Synnax Console by connecting the press_pt channel
to a line plot and making a schematic that shows valves representing press_vlv_cmd and
vent_vlv_cmd and a tank with a value representing press_pt.
"""

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Define the control channel names
PRESS_VALVE = "press_vlv_cmd"
VENT_VALVE = "vent_vlv_cmd"
PRESSURE = "press_pt"

# Open a control sequence under a context manager, so that the control is released when
# the block exits
with client.control.acquire(
    # A useful name that identifies the sequence to the rest of the system. We highly
    # recommend keeping these names unique across your sequences.
    name="Pressurization Sequence",
    # Defines the authorities at which the sequence controls the valve channels. This is
    # a number from 0 to 255. A writer with a higher control authority can override a
    # writer with a lower control control authority.
    write_authorities=[200],
    # We need to set the channels we'll be writing to and reading from.
    write=[PRESS_VALVE, VENT_VALVE],
    read=[PRESSURE],
) as ctrl:
    # Mark the start of the sequence
    start = sy.TimeStamp.now()

    # Set the initial target pressure
    target_pressure = 20

    # Close the vent valve
    ctrl[VENT_VALVE] = False

    # Pressurize the system five times in 20 psi increments
    for i in range(5):
        # Open the pressurization valve
        ctrl[PRESS_VALVE] = True
        if ctrl.wait_until(
            # Wait until the pressure is greater than the current target
            lambda c: c[PRESSURE] > target_pressure,
            # If the pressure doesn't reach the target in 20 seconds, break the loop and
            # vent the system
            timeout=20 * sy.TimeSpan.SECOND,
        ):
            # Close the pressurization valve
            ctrl[PRESS_VALVE] = False
            # Wait for 2 seconds
            ctrl.sleep(2)
            # Increment the target
            target_pressure += 20
        else:
            break

    # Vent the system
    ctrl[VENT_VALVE] = True

    # Wait until the pressure is less than 5 psi
    ctrl.wait_until(lambda c: c[PRESSURE] < 5)

    # Close the vent valve
    ctrl[VENT_VALVE] = False

    # Mark the end of the sequence
    end = sy.TimeStamp.now()

    # Label the sequence with the end time
    client.ranges.create(
        name=f"Auto Pressurization Sequence {end}",
        time_range=sy.TimeRange(start=start, end=end),
    )



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TypeAlias, cast, overload

from pandas import DataFrame
from pydantic import BaseModel, Field

import synnax.channel.payload as channel
from synnax.exceptions import ValidationError
from synnax.telem import CrudeSeries, MultiSeries, Series, TypedCrudeSeries


class FramePayload(BaseModel):
    keys: list[channel.Key]
    series: list[Series]

    def __init__(
        self,
        keys: list[int] | None = None,
        series: list[Series] | None = None,
    ):
        if series is None:
            series = list()
        if keys is None:
            keys = list()
        super().__init__(series=series, keys=keys)


class Frame:
    """
    A frame is a collection of telemetry series mapped to particular channels. Frames
    can be keyed by channel name or channel key, but not both.
    """

    channels: list[channel.Key] | list[str]
    series: list[Series] = Field(default_factory=list)

    def __init__(
        self,
        channels: (
            list[channel.Key]
            | tuple[channel.Key]
            | list[str]
            | tuple[str]
            | DataFrame
            | Frame
            | FramePayload
            | dict[channel.Key, TypedCrudeSeries]
            | dict[str, TypedCrudeSeries]
            | None
        ) = None,
        series: Sequence[TypedCrudeSeries] | None = None,
    ):
        if isinstance(channels, Frame):
            self.channels = channels.channels
            self.series = channels.series
        elif isinstance(channels, FramePayload):
            self.channels = list(channels.keys)
            self.series = channels.series
        elif isinstance(channels, DataFrame):
            self.channels = channels.columns.to_list()
            self.series = [Series(data=channels[k]) for k in self.channels]
        elif isinstance(channels, dict):
            self.channels = cast(list[channel.Key] | list[str], list(channels.keys()))
            self.series = [Series(d) for d in channels.values()]
        elif (series is None or isinstance(series, list)) and (
            channels is None or isinstance(channels, list)
        ):
            self.series = list() if series is None else [Series(d) for d in series]
            self.channels = (
                cast(list[channel.Key] | list[str], list(channels))
                if channels
                else list()
            )
        else:
            raise ValueError(f"""
                [Frame] - invalid construction arguments. Received {channels}
                and {series}.
            """)

    def __str__(self) -> str:
        return self.to_df().__str__()

    @overload
    def append(self, col_or_frame: channel.Key | str, series: Series) -> None:
        """Adds a new series to the frame for the given channel.
        :param col_or_frame: The channel key or name to append the series to.
        :param series: The series to append to the frame
        """
        ...

    @overload
    def append(self, col_or_frame: Frame) -> None:
        """Appends the given frame to the current frame, modifying the current frame in place.
        :param col_or_frame: The frame to append to the current frame.
        """
        ...

    def append(
        self,
        col_or_frame: channel.Key | str | Frame,
        series: Series | None = None,
    ) -> None:
        """Appends a frame or series to the current frame. If a frame is provided, the
        series and channels from the frame are appended to the current frame. If a series
        is provided, the series is appended to the frame for the given channel.
        :param col_or_frame: The channel key or name to append the series to, or the frame
            to append to the current frame.
        :param series: The series to append to the frame, if col_or_frame is a channel key
            or name.
        """
        if isinstance(col_or_frame, Frame):
            self.series.extend(col_or_frame.series)
            self.channels.extend(col_or_frame.channels)  # type: ignore[arg-type]
        else:
            if series is None:
                raise ValueError("series must be provided when appending a channel")
            self.series.append(series)
            self.channels.append(col_or_frame)  # type: ignore[arg-type]

    def items(
        self,
    ) -> Iterator[tuple[channel.Key | str, Series]]:
        """
        Returns a generator of tuples containing the channel and series for each channel
        in the frame.
        """
        return zip(self.channels, self.series)

    def __getitem__(self, key: channel.Key | str) -> MultiSeries:
        if not isinstance(key, (channel.Key, str)):
            return self.to_df()[key]
        indexes = [i for i, k in enumerate(self.channels) if k == key]
        return MultiSeries([self.series[i] for i in indexes])

    def get(
        self, key: channel.Key | str, default: MultiSeries | None = None
    ) -> MultiSeries | None:
        """Gets the series for the given channel key or name. If the channel does not
        exist in the frame, returns the default value or None if no default is provided.
        :param key: The channel key or name to get the series for.
        :param default: The default value to return if the channel does not exist in the
        """
        try:
            return self[key]
        except ValueError:
            return default

    def to_payload(self) -> FramePayload:
        """Converts the frame to its payload representation for transport over the
        network. This method should typically only be used internally.
        :raises: ValidationError if the frame is keyed by channel name instead of key.
        """
        if not all(isinstance(k, channel.Key) for k in self.channels):
            diff = [k for k in self.channels if not isinstance(k, channel.Key)]
            raise ValidationError(f"""
            Cannot convert a frame that contains channel names to a payload.
            The following channels are invalid: {diff}
            """)
        keys: list[int] = [k for k in self.channels if isinstance(k, int)]
        return FramePayload(keys=keys, series=self.series)

    def to_df(self) -> DataFrame:
        """Converts the frame to a pandas DataFrame. Each column in the DataFrame
        corresponds to a channel in the frame.
        """
        base: dict[channel.Key | str, object] = dict()
        channels = cast(list[channel.Key | str], self.channels)
        for k in set(channels):
            v = self.get(k)
            if v is None:
                continue
            try:
                base[k] = v.__array__()
            except TypeError:
                base[k] = list(v)
        return DataFrame(base)

    def __contains__(self, key: channel.Key | str) -> bool:
        return key in self.channels


CrudeFrame: TypeAlias = (
    Frame | FramePayload | dict[channel.Key, CrudeSeries] | DataFrame
)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from abc import ABC, abstractmethod

import synnax as sy


class Simulator(ABC):
    """Base class for all simulators."""

    def __init__(self, verbose: bool = False):
        self.verbose = verbose
        self._running = False

    def log(self, message: str) -> None:
        """Print message only when verbose mode is enabled."""
        if self.verbose:
            print(f"[{self.__class__.__name__}] {message}")

    @abstractmethod
    def start(self) -> None:
        """Start the simulator."""
        ...

    @abstractmethod
    def stop(self, timeout: sy.TimeSpan = 5 * sy.TimeSpan.SECOND) -> None:
        """Stop the simulator."""
        ...



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example demonstrates how to stream live data from a channel in Synnax.
Live-streaming is useful for real-time data processing and analysis, and is an
integral part of Synnax's control sequence and data streaming capabilities.

This example is meant to be used in conjunction with the `stream_write.py` example, and
assumes that example is running in a separate terminal.
"""

import synnax as sy

# We've logged in via the CLI, so there's no need to provide credentials here.
# See https://docs.synnaxlabs.com/reference/client/quick-start for more information.
client = sy.Synnax()

# We can just specify the names of the channels we'd like to stream from.
read_from = [
    "stream_write_example_time",
    "stream_write_example_data_1",
    "stream_write_example_data_2",
]

# Open the streamer as a context manager. This will make sure the streamer is properly
# closed when we're done reading. We'll read from both the time and data channels. In
with client.open_streamer(read_from) as s:
    # Loop through the frames in the streamer. Each iteration will block until a new
    # frame is available, then we'll just print it out.
    over_threshold = False
    start = None
    for frame in s:
        if frame["stream_write_example_data_1"][0] > 0 and not over_threshold:
            print("Exceeds Threshold")
            over_threshold = True
            start = frame["stream_write_example_time"][0]
            print(over_threshold)
        elif over_threshold and frame["stream_write_example_data_1"][0] < 0:
            print("Data Exceeds Threshold")
            end = frame["stream_write_example_time"][0]
            tr = sy.TimeRange(start, end)
            client.ranges.create(
                name="Data Exceeds Threshold",
                time_range=tr,
                color="#BA55D3",
            )
            over_threshold = False



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import threading
from uuid import uuid4

import pytest

import synnax as sy


@pytest.mark.task
class TestTaskClient:
    def test_create_single(self, client: sy.Synnax):
        task = client.tasks.create(name="test", type="test")
        assert task.key != 0

    def test_create_multiple(self, client: sy.Synnax):
        t1 = sy.Task(name="test1", type="test")
        t2 = sy.Task(name="test2", type="test")
        tasks = client.tasks.create(tasks=[t1, t2])
        assert len(tasks) == 2
        assert tasks[0].name == "test1"
        assert tasks[1].name == "test2"

    def test_retrieve_by_name(self, client: sy.Synnax):
        name = str(uuid4())
        task = client.tasks.create(name=name, type="test")
        res = client.tasks.retrieve(name=name)
        assert res.name == name
        assert res.key == task.key

    def test_retrieve_by_type(self, client: sy.Synnax):
        type = str(uuid4())
        task = client.tasks.create(type=type)
        res = client.tasks.retrieve(type=type)
        assert res.type == type
        assert res.key == task.key

    def test_execute_command_sync(self, client: sy.Synnax):
        def driver(ev: threading.Event):
            with client.open_streamer("sy_task_cmd") as s:
                ev.set()
                f = s.read(timeout=1)
                cmd = f["sy_task_cmd"][0]
                client.statuses.set(
                    sy.Status(
                        key=str(sy.task.payload.ontology_id(cmd["task"])),
                        variant=sy.status.VARIANT_SUCCESS,
                        message="Command executed.",
                        details=sy.task.StatusDetails(
                            task=int(cmd["task"]), cmd=cmd["key"]
                        ),
                    )
                )

        ev = threading.Event()
        t = threading.Thread(target=driver, args=(ev,))
        t.start()
        tsk = client.tasks.create(name="test", type="test")
        ev.wait()
        tsk.execute_command_sync("test", {"key": "value"})
        t.join()

    def test_task_configure_success(self, client: sy.Synnax):
        """Should not throw an error when the task is configured successfully."""

        def driver(ev: threading.Event):
            with client.open_streamer("sy_task_set") as s:
                ev.set()
                f = s.read(timeout=2)
                key = f["sy_task_set"][0]
                client.statuses.set(
                    sy.Status(
                        key=str(sy.task.payload.ontology_id(int(key))),
                        variant=sy.status.VARIANT_SUCCESS,
                        message="Task configured.",
                        details=sy.task.StatusDetails(task=int(key)),
                    )
                )

        tsk = sy.Task()
        ev = threading.Event()
        t = threading.Thread(target=driver, args=(ev,))
        t.start()
        ev.wait()
        client.tasks.configure(tsk)
        t.join()

    def test_task_configure_invalid_config(self, client: sy.Synnax):
        """Should throw an error when the driver responds with an error"""

        def driver(ev: threading.Event):
            with client.open_streamer("sy_task_set") as s:
                ev.set()
                f = s.read(timeout=1)
                key = f["sy_task_set"][0]
                client.statuses.set(
                    sy.Status(
                        key=str(sy.task.payload.ontology_id(int(key))),
                        variant=sy.status.VARIANT_ERROR,
                        message="Invalid Configuration.",
                        details=sy.task.StatusDetails(task=int(key)),
                    )
                )

        tsk = sy.Task()
        ev = threading.Event()
        t = threading.Thread(target=driver, args=(ev,))
        t.start()
        ev.wait()
        with pytest.raises(sy.ConfigurationError, match="Invalid Configuration."):
            client.tasks.configure(tsk)
        t.join()

    def test_task_configure_timeout(self, client: sy.Synnax):
        """Should throw an error when the task is not configured within the timeout."""
        tsk = sy.Task()
        with pytest.raises(TimeoutError):
            client.tasks.configure(tsk, timeout=0.1)

    def test_list_tasks(self, client: sy.Synnax):
        """Should list all tasks on the default rack."""
        # Create some tasks
        task1 = client.tasks.create(name=str(uuid4()), type="test1")
        task2 = client.tasks.create(name=str(uuid4()), type="test2")

        # List all tasks
        tasks = client.tasks.list()

        # Should contain at least the tasks we just created
        task_keys = [t.key for t in tasks]
        assert task1.key in task_keys
        assert task2.key in task_keys

    def test_copy_task(self, client: sy.Synnax):
        """Should copy a task with a new name."""
        # Create an original task
        original_name = str(uuid4())
        original = client.tasks.create(
            name=original_name, type="test", config={"foo": "bar"}
        )

        # Copy the task
        copy_name = str(uuid4())
        copied = client.tasks.copy(
            key=original.key,
            name=copy_name,
        )

        # Verify the copy
        assert copied.key != original.key
        assert copied.name == copy_name
        assert copied.type == original.type
        assert copied.config == original.config



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import Protocol

from alamos import NOOP, Instrumentation, trace
from freighter import UnaryClient, send_required
from pydantic import BaseModel

from synnax.channel.payload import (
    Key,
    NormalizedNameResult,
    Params,
    Payload,
    normalize_params,
)
from synnax.exceptions import NotFoundError


class _Request(BaseModel):
    names: list[str] | None = None
    keys: list[int] | None = None
    leaseholder: int | None = None


class _Response(BaseModel):
    channels: list[Payload] = list()
    not_found: list[str] | None = list()


class Retriever(Protocol):
    """Protocol for retrieving channel payloads from the cluster."""

    def retrieve(self, channels: Params) -> list[Payload]: ...

    def retrieve_one(self, param: Key | str) -> Payload: ...


class ClusterRetriever:
    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self.instrumentation = instrumentation

    def _(self) -> Retriever:
        return self

    @trace("debug")
    def retrieve(self, channels: Params) -> list[Payload]:
        normal = normalize_params(channels)
        if len(normal.channels) == 0:
            return list()
        if isinstance(normal, NormalizedNameResult):
            req = _Request(names=normal.channels)
        else:
            req = _Request(keys=normal.channels)
        return self.__exec_retrieve(req)

    @trace("debug")
    def retrieve_one(self, param: Key | str) -> Payload:
        req = _Request()
        if isinstance(param, Key):
            req.keys = [param]
        else:
            req.names = [param]
        channels = self.__exec_retrieve(req)
        if len(channels) == 0:
            raise NotFoundError(f"Could not find channel matching {param}")
        return channels[0]

    def __exec_retrieve(self, req: _Request) -> list[Payload]:
        return send_required(self._client, "/channel/retrieve", req, _Response).channels


class CacheRetriever:
    _retriever: Retriever
    _channels: dict[Key, Payload]
    _names_to_keys: dict[str, set[Key]]
    instrumentation: Instrumentation

    def __init__(
        self,
        retriever: Retriever,
        instrumentation: Instrumentation,
    ) -> None:
        self._channels = dict()
        self._names_to_keys = dict()
        self.instrumentation = instrumentation
        self._retriever = retriever

    def delete(self, keys: Params) -> None:
        normal = normalize_params(keys)
        if isinstance(normal, NormalizedNameResult):
            matches = {
                ch for ch in self._channels.values() if ch.name in normal.channels
            }
            for ch in matches:
                self._channels.pop(ch.key)
                self._names_to_keys.pop(ch.name)
        else:
            for key in normal.channels:
                channel = self._channels.get(key)
                if channel is not None:
                    self._channels.pop(key)
                    self._names_to_keys.pop(channel.name)

    def rename(self, keys: list[Key], names: list[str]) -> None:
        for key, name in zip(keys, names):
            channel = self._channels.get(key)
            if channel is None:
                continue
            self._channels.pop(key)
            existing_keys = self._names_to_keys.get(channel.name)
            if existing_keys is not None:
                existing_keys.remove(key)
            channel.name = name
            self._channels[channel.key] = channel
            existing_keys = self._names_to_keys.get(name)
            if existing_keys is None:
                self._names_to_keys[name] = {channel.key}
            else:
                existing_keys.add(channel.key)

    def _(self) -> Retriever:
        return self

    def _get(self, param: Key | str) -> list[Payload] | None:
        if isinstance(param, Key):
            ch = self._channels.get(param)
            return [ch] if ch is not None else None
        keys = self._names_to_keys.get(param, set())
        channels = list()
        for key in keys:
            ch = self._channels.get(key)
            if ch is not None:
                channels.append(ch)
        return None if len(channels) == 0 else channels

    def _get_one(self, param: Key | str) -> Payload | None:
        if isinstance(param, Key):
            return self._channels.get(param)
        keys = self._names_to_keys.get(param, None)
        if keys is None:
            return None
        return self._channels.get(next(iter(keys)))

    def set(self, channels: list[Payload]) -> None:
        for channel in channels:
            self._set_one(channel)

    def _set_one(self, channel: Payload) -> None:
        self._channels[channel.key] = channel
        keys = self._names_to_keys.get(channel.name)
        if keys is None:
            self._names_to_keys[channel.name] = {channel.key}
        else:
            keys.add(channel.key)

    @trace("debug")
    def retrieve(self, channels: Params) -> list[Payload]:
        normal = normalize_params(channels)
        results: list[Payload] = []
        missed: list[int] = []
        if isinstance(normal, NormalizedNameResult):
            params: list[Key | str] = list(normal.channels)
        else:
            params = list(normal.channels)
        for i, param in enumerate(params):
            ch = self._get(param)
            if ch is None:
                missed.append(i)
            else:
                results.extend(ch)
        if not missed:
            return results
        if isinstance(normal, NormalizedNameResult):
            to_retrieve: Params = [normal.channels[i] for i in missed]
        else:
            to_retrieve = [normal.channels[i] for i in missed]
        retrieved = self._retriever.retrieve(to_retrieve)
        self.set(retrieved)
        results.extend(retrieved)
        return results

    def retrieve_one(self, param: Key | str) -> Payload:
        ch = self._get_one(param)
        if ch is not None:
            return ch
        retrieved = self._retriever.retrieve_one(param)
        self._set_one(retrieved)
        return retrieved


def retrieve_required(r: Retriever, channels: Params) -> list[Payload]:
    normal = normalize_params(channels)
    results = r.retrieve(channels)
    found: set[Key | str]
    if isinstance(normal, NormalizedNameResult):
        found = {c.name for c in results}
    else:
        found = {c.key for c in results}
    not_found: list[Key | str] = [p for p in normal.channels if p not in found]
    if len(not_found) > 0:
        raise NotFoundError(f"Could not find channels: {not_found}")
    return results



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import uuid4

import pytest
from freighter.mock import MockUnaryClient

import synnax as sy


@pytest.mark.auth
class TestClusterAuth:
    def test_invalid_credentials(self, login_info: tuple[str, int, str, str]):
        """
        Should raise an AuthError
        """
        host, port, username, _ = login_info
        with pytest.raises(sy.AuthError):
            sy.Synnax(
                host=host,
                port=port,
                username=username,
                password="wrong",
            )

    def test_no_password_provided(self, login_info: tuple[str, int, str, str]):
        """
        Should raise a ValidationError
        """
        host, port, username, _ = login_info
        with pytest.raises(sy.PathError):
            sy.Synnax(
                host=host,
                port=port,
                username=username,
            )


@pytest.mark.auth
class TestAuthRetry:
    @pytest.fixture(scope="function")
    def auth_setup(self) -> MockUnaryClient[int, int]:
        """Fixture that provides the mock clients and auth setup for retry tests."""
        # Setup mock login client with two successful login responses
        res = sy.auth.TokenResponse(
            token="abc",
            user=sy.User(
                key=uuid4(),
                username="synnax",
                password="seldon",
                email="synnax@synnax.com",
                first_name="Synnax",
                last_name="Labs",
            ),
        )
        mock_login_client = MockUnaryClient[
            sy.auth.InsecureCredentials, sy.auth.TokenResponse
        ](responses=[res, res], response_errors=[None, None])

        # Create auth client
        auth = sy.auth.Client(mock_login_client, "synnax", "seldon")

        # Create base mock client
        mock_client = MockUnaryClient[int, int](responses=[1, 1])
        mock_client.use(auth.middleware())

        return mock_client

    def test_retry_on_invalid_token(self, auth_setup: MockUnaryClient[int, int]):
        """Test that authentication retries when receiving an invalid token error."""
        mock_client = auth_setup
        mock_client.response_errors = [sy.InvalidToken("invalid token"), None]

        response, error = mock_client.send("", 1, int)
        assert error is None
        assert response == 1

    def test_retry_on_expired_token(self, auth_setup):
        """Test that authentication retries when receiving an expired token error."""
        mock_client = auth_setup
        mock_client.response_errors = [sy.ExpiredToken("token expired"), None]
        response, error = mock_client.send("", 1, int)
        assert error is None
        assert response == 1



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import numpy as np

import synnax as sy

# Construct a series from a list of numbers. In this case, the series will
# automatically be of type int64.
series = sy.Series([1, 2, 3, 4, 5])

# Construct a series from a list of numbers, but this time we specify the type
# explicitly.
series = sy.Series([1, 2, 3, 4, 5], data_type=sy.DataType.FLOAT32)

# Construct a series from a list of strings. In this case, the series will
# automatically be of type string.
series = sy.Series(["apple", "banana", "cherry"])

# Construct a series from a numpy array. This is the most efficient way to construct a
# series from a large amount of data.
series = sy.Series(np.array([1, 2, 3, 4, 5], dtype=np.float32))

# Construct a series from a list of JSON objects. This is useful when you have a series
# that has been serialized to JSON.
series = sy.Series([{"red": "cherry"}, {"yellow": "banana"}, {"orange": "orange"}])

series = sy.Series([1, 2, 3, 4, 5])

print(series[0])  # 1
print(series[-1])  # 5

series = sy.Series([1, 2, 3, 4, 5])
# Access the underlying numpy array
arr = np.array(series)
print(arr)  # [1 2 3 4 5]

series = sy.Series([1, 2, 3, 4, 5])
py_list = list(series)
print(py_list)  # [1, 2, 3, 4, 5]

start = sy.TimeStamp.now()

tr = sy.TimeRange(start, start + 5 * sy.TimeSpan.SECOND)

series = sy.Series(
    data=[1, 2, 3, 4, 5],
    data_type=sy.DataType.FLOAT64,
    time_range=tr,
)

series = sy.Series([1, 2, 3, 4, 5])
print(len(series))  # 5

string_series = sy.Series(["apple", "banana", "cherry"])
print(len(string_series))  # 3

series = sy.Series([1, 2, 3, 4, 5])
print(series.data_type)  # int64
print(series.data_type == sy.DataType.STRING)  # False

series = sy.Series([1, 2, 3, 4, 5])
print(max(series))  # 5
print(min(series))  # 1

# Construct a frame for the given channel names.
frame = sy.Frame(
    {
        "channel1": sy.Series([1, 2, 3, 4, 5]),
        "channel2": sy.Series([5, 4, 3, 2, 1]),
        "channel3": sy.Series([1, 1, 1, 1, 1]),
    }
)

# Construct a frame for the given channel keys
frame = sy.Frame(
    {
        1: sy.Series([1, 2, 3, 4, 5]),
        2: sy.Series([5, 4, 3, 2, 1]),
        # Notice that series do not need to be the same length.
        3: sy.Series([1, 1, 1]),
    }
)

# Or construct a frame with multiple series for a single channel
frame = sy.Frame(
    {
        "channel1": [
            sy.Series([1, 2, 3, 4, 5]),
            sy.Series([5, 4, 3, 2, 1]),
            sy.Series([1, 1, 1, 1, 1]),
        ],
        "channel2": [sy.Series([1, 2, 3, 4, 5])],
    }
)

frame = sy.Frame(
    {
        "channel1": [sy.Series([1, 2]), sy.Series([3, 4, 5])],
        "channel2": sy.Series([5, 4, 3, 2, 1]),
        "channel3": sy.Series([1, 1, 1, 1, 1]),
    }
)

multi_series = frame["channel1"]
# Access a value
print(multi_series[0])  # 1

# Access a value from a specific series
print(multi_series.series[0][0])  # 1

# Construct a Python list from the MultiSeries
py_list = list(multi_series)
print(py_list)  # [1, 2, 3, 4, 5]

frame = sy.Frame(
    {
        "channel1": sy.Series([1, 2, 3, 4, 5]),
        "channel2": sy.Series([5, 4, 3, 2, 1]),
        "channel3": sy.Series([1, 1]),
    }
)

obj = frame[3]
print(obj)  # {'channel1': 4, 'channel2': 2, 'channel3': None}

# Convert the frame to a pandas DataFrame
df = frame.to_df()
print(df)



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from pydantic import BaseModel, model_validator


class Color(BaseModel):
    """An RGBA color with 8-bit RGB channels and a float alpha."""

    r: int = 0
    g: int = 0
    b: int = 0
    a: float = 1

    def __init__(self, __value: object = None, /, **kwargs: object):
        if __value is not None:
            validated = type(self).model_validate(__value)
            super().__init__(r=validated.r, g=validated.g, b=validated.b, a=validated.a)
        else:
            super().__init__(**kwargs)

    @model_validator(mode="before")
    @classmethod
    def _parse(cls, v: object) -> object:
        if isinstance(v, Color):
            return v
        if isinstance(v, str):
            return _from_hex(v)
        if isinstance(v, (list, tuple)):
            if len(v) == 3:
                return {"r": int(v[0]), "g": int(v[1]), "b": int(v[2]), "a": 1.0}
            if len(v) == 4:
                return {
                    "r": int(v[0]),
                    "g": int(v[1]),
                    "b": int(v[2]),
                    "a": float(v[3]),
                }
            raise ValueError(f"Invalid color array length: {len(v)}")
        if isinstance(v, dict):
            return v
        raise ValueError(f"Cannot parse color from: {v!r}")

    def hex(self) -> str:
        """Return the hex string representation of the color."""
        alpha_byte = round(self.a * 255)
        if alpha_byte == 255:
            return f"#{self.r:02x}{self.g:02x}{self.b:02x}"
        return f"#{self.r:02x}{self.g:02x}{self.b:02x}{alpha_byte:02x}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (str, list, tuple)):
            try:
                other = Color(other)
            except (ValueError, TypeError):
                return NotImplemented
        return super().__eq__(other)

    @property
    def is_zero(self) -> bool:
        return self.r == 0 and self.g == 0 and self.b == 0 and self.a == 0


def _from_hex(s: str) -> dict[str, int | float]:
    s = s.lstrip("#")
    if len(s) == 0:
        return {"r": 0, "g": 0, "b": 0, "a": 0}
    if len(s) == 6:
        return {
            "r": int(s[0:2], 16),
            "g": int(s[2:4], 16),
            "b": int(s[4:6], 16),
            "a": 1.0,
        }
    if len(s) == 8:
        return {
            "r": int(s[0:2], 16),
            "g": int(s[2:4], 16),
            "b": int(s[4:6], 16),
            "a": int(s[6:8], 16) / 255.0,
        }
    raise ValueError(f"Invalid hex color: #{s}")


Crude = Color | str | list[int] | tuple[int, ...]
"""Types that can be coerced into a Color."""



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from .device_sim import DeviceSim
from .load_current import LoadCurrentSimDAQ
from .press import PressSimDAQ
from .simdaq import SimDAQ
from .simulator import Simulator
from .thermal import ThermalSimDAQ
from .tpc import TPCSimDAQ

__all__ = [
    "Simulator",
    "SimDAQ",
    "DeviceSim",
    "LoadCurrentSimDAQ",
    "PressSimDAQ",
    "ThermalSimDAQ",
    "TPCSimDAQ",
]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.opcua.types import (
    Channel,
    Device,
    ReadChannel,
    ReadTask,
    SecurityMode,
    SecurityPolicy,
    WrappedReadTaskConfig,
    WriteChannel,
    WriteTask,
    WriteTaskConfig,
)

__all__ = [
    "Channel",
    "Device",
    "ReadChannel",
    "ReadTask",
    "SecurityMode",
    "SecurityPolicy",
    "WrappedReadTaskConfig",
    "WriteChannel",
    "WriteTask",
    "WriteTaskConfig",
]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.labjack.types import (
    SUPPORTED_MODELS,
    T4,
    T7,
    T8,
    AIChan,
    Device,
    DIChan,
    OutputChan,
    ReadTask,
    ReadTaskConfig,
    ThermocoupleChan,
    WriteTask,
    WriteTaskConfig,
)

__all__ = [
    "AIChan",
    "Device",
    "DIChan",
    "OutputChan",
    "ReadTask",
    "ReadTaskConfig",
    "SUPPORTED_MODELS",
    "T4",
    "T7",
    "T8",
    "ThermocoupleChan",
    "WriteTask",
    "WriteTaskConfig",
]



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

"""
This example shows how to create a named range in Synnax, which can be used to identify
and lookup specific periods of time in your data.

We'll write  data to an index and data channel, and then create a range that spans the
entire time range of the data. Then, we'll show how to read the data back using the
range.
"""

import matplotlib.pyplot as plt
import numpy as np

import synnax as sy

# We've logged in via the command-line interface, so there's no need to provide
# credentials here. See https://docs.synnaxlabs.com/reference/client/quick-start.
client = sy.Synnax()

# Define the data.
start = sy.TimeStamp.now()
end = start + 10 * sy.TimeSpan.SECOND

time_data = np.linspace(start, end, 1000)
data = np.sin(time_data - start)

# Create an index channel that will be used to store our timestamps.
time_channel = client.channels.create(
    name="create_range_time",
    data_type=sy.DataType.TIMESTAMP,
    is_index=True,
)

# Create a data channel that will be used to store our fake sensor data.
data_channel = client.channels.create(
    name="create_range_data",
    data_type=sy.DataType.FLOAT64,
    index=time_channel.key,
)

# Write the data to the Synnax cluster through the channels. Note that we need to write
# to the index channel first, otherwise our write will fail.
time_channel.write(start, time_data)
data_channel.write(start, data)

# Create a range that spans the start and end of the data.
example_range = client.ranges.create(
    name="create_range_range",
    time_range=sy.TimeRange(start, end),
)

# We can pull and plot the data from the range by just accessing the channel names as if
# they were attributes of the range itself.
plt.plot(
    # The elapsed_seconds method converts the timestamps to seconds since the start of
    # the range.
    sy.elapsed_seconds(example_range["create_range_time"]),
    example_range["create_range_data"],
)
plt.xlabel("Time")
plt.ylabel("Value")
plt.show()



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import time
from uuid import uuid4

import pytest

import synnax as sy


@pytest.mark.arc
class TestArc:
    """Tests for Arc CRUD operations."""

    def test_create_from_kwargs(self, client: sy.Synnax):
        """Should create an Arc from keyword arguments."""
        name = f"test-arc-{uuid4()}"
        arc = client.arcs.create(
            name=name,
            mode="text",
            text=sy.Text(raw="x := 1"),
        )
        assert arc.name == name
        assert arc.key != sy.Arc().key
        assert arc.mode == "text"
        assert arc.text.raw == "x := 1"

    def test_create_from_instance(self, client: sy.Synnax):
        """Should create an Arc from an existing instance."""
        name = f"test-arc-{uuid4()}"
        arc = sy.Arc(
            name=name,
            mode="text",
            text=sy.Text(raw="y := 2"),
        )
        created = client.arcs.create(arc)
        assert created.name == name
        assert created.key != sy.Arc().key
        assert created.text.raw == "y := 2"

    def test_create_multiple(self, client: sy.Synnax):
        """Should create multiple Arcs at once."""
        arcs = [
            sy.Arc(name=f"test-arc-multi-{uuid4()}", mode="text"),
            sy.Arc(name=f"test-arc-multi-{uuid4()}", mode="text"),
        ]
        created = client.arcs.create(arcs)
        assert len(created) == 2
        for arc in created:
            assert arc.name.startswith("test-arc-multi-")
            assert arc.key != sy.Arc().key

    def test_retrieve_by_key(self, client: sy.Synnax):
        """Should retrieve an Arc by key."""
        name = f"test-arc-{uuid4()}"
        created = client.arcs.create(name=name, mode="text")
        retrieved = client.arcs.retrieve(key=created.key)
        assert retrieved.key == created.key
        assert retrieved.name == name

    def test_retrieve_by_name(self, client: sy.Synnax):
        """Should retrieve an Arc by name."""
        name = f"test-arc-{uuid4()}"
        created = client.arcs.create(name=name, mode="text")
        retrieved = client.arcs.retrieve(name=name)
        assert retrieved.key == created.key
        assert retrieved.name == name

    def test_retrieve_multiple_by_keys(self, client: sy.Synnax):
        """Should retrieve multiple Arcs by keys."""
        arcs = [
            sy.Arc(name=f"test-arc-keys-{uuid4()}", mode="text"),
            sy.Arc(name=f"test-arc-keys-{uuid4()}", mode="text"),
        ]
        created = client.arcs.create(arcs)
        keys = [arc.key for arc in created]
        retrieved = client.arcs.retrieve(keys=keys)
        assert len(retrieved) == 2
        retrieved_keys = {arc.key for arc in retrieved}
        assert retrieved_keys == set(keys)

    def test_retrieve_not_found(self, client: sy.Synnax):
        """Should raise NotFoundError when Arc is not found."""
        with pytest.raises(sy.NotFoundError):
            client.arcs.retrieve(name=f"nonexistent-arc-{uuid4()}")

    def test_delete_single(self, client: sy.Synnax):
        """Should delete a single Arc."""
        name = f"test-arc-delete-{uuid4()}"
        arc = client.arcs.create(name=name, mode="text")
        client.arcs.delete(arc.key)
        with pytest.raises(sy.NotFoundError):
            client.arcs.retrieve(key=arc.key)

    def test_delete_multiple(self, client: sy.Synnax):
        """Should delete multiple Arcs."""
        arcs = [
            sy.Arc(name=f"test-arc-delete-multi-{uuid4()}", mode="text"),
            sy.Arc(name=f"test-arc-delete-multi-{uuid4()}", mode="text"),
        ]
        created = client.arcs.create(arcs)
        keys = [arc.key for arc in created]
        client.arcs.delete(keys)
        for key in keys:
            with pytest.raises(sy.NotFoundError):
                client.arcs.retrieve(key=key)

    def test_search(self, client: sy.Synnax):
        """Should search for Arcs by name."""
        unique_prefix = f"searchable-arc-{uuid4()}"
        client.arcs.create(name=f"{unique_prefix}-1", mode="text")
        client.arcs.create(name=f"{unique_prefix}-2", mode="text")
        time.sleep(0.2)
        results = client.arcs.retrieve(search_term=unique_prefix)
        assert len(results) >= 2
        for arc in results:
            assert unique_prefix in arc.name


@pytest.mark.arc
class TestArcGraph:
    """Tests for Arc graph mode functionality."""

    def test_graph_with_edges(self, client: sy.Synnax):
        """Should create an Arc with graph nodes and edges."""
        name = f"test-arc-graph-{uuid4()}"
        graph = sy.Graph(
            nodes=[
                sy.GraphNode(
                    key="node1",
                    type="constant",
                    config={"value": 1},
                    position=sy.Position(x=0, y=0),
                ),
                sy.GraphNode(
                    key="node2",
                    type="add",
                    config={},
                    position=sy.Position(x=100, y=0),
                ),
            ],
            edges=[
                sy.Edge(
                    source=sy.Handle(param="output", node="node1"),
                    target=sy.Handle(param="input1", node="node2"),
                ),
            ],
        )
        arc = client.arcs.create(name=name, mode="graph", graph=graph)
        assert arc.mode == "graph"
        assert len(arc.graph.nodes) == 2
        assert len(arc.graph.edges) == 1
        assert arc.graph.nodes[0].key == "node1"
        assert arc.graph.edges[0].source.node == "node1"
        assert arc.graph.edges[0].target.node == "node2"


@pytest.mark.arc
class TestArcTask:
    """Tests for Arc task functionality."""

    def test_create_arc_task(self, client: sy.Synnax):
        """Should create an Arc task."""
        arc = client.arcs.create(
            name=f"test-arc-task-{uuid4()}",
            mode="text",
            text=sy.Text(raw="x := 1"),
        )
        task = sy.arc.Task(name="Test Arc Task", arc_key=arc.key, auto_start=False)
        assert task.config.arc_key == str(arc.key)
        assert task.config.auto_start is False
        assert task.TYPE == "arc"

    def test_arc_task_to_payload(self, client: sy.Synnax):
        """Should convert Arc task to payload."""
        arc = client.arcs.create(
            name=f"test-arc-task-{uuid4()}",
            mode="text",
        )
        task = sy.arc.Task(name="Test Arc Task", arc_key=arc.key)
        payload = task.to_payload()
        assert payload.type == "arc"
        assert payload.name == "Test Arc Task"
        assert payload.config["arc_key"] == str(arc.key)

    def test_arc_task_requires_arc_key(self):
        """Should raise error when arc_key is not provided."""
        with pytest.raises(ValueError, match="arc_key is required"):
            sy.arc.Task(name="Test")

    def test_create_arc_task_on_cluster(self, client: sy.Synnax):
        """Should create an Arc task on the cluster."""
        arc = client.arcs.create(
            name=f"test-arc-task-cluster-{uuid4()}",
            mode="text",
            text=sy.Text(raw="x := 1"),
        )
        task_name = f"Test Arc Task {uuid4()}"
        task = sy.arc.Task(name=task_name, arc_key=arc.key, auto_start=False)
        # Create task using to_payload pattern (matching modbus/opcua tests)
        created = client.tasks.create(
            name=task_name,
            type="arc",
            config=task.to_payload().config,
        )
        assert created.key != 0
        assert created.name == task_name
        assert created.type == "arc"
        assert created.config["arc_key"] == str(arc.key)

    def test_retrieve_arc_task(self, client: sy.Synnax):
        """Should retrieve an Arc task from the cluster."""
        arc = client.arcs.create(
            name=f"test-arc-task-retrieve-{uuid4()}",
            mode="text",
        )
        task_name = f"Test Arc Task {uuid4()}"
        task = sy.arc.Task(name=task_name, arc_key=arc.key)
        created = client.tasks.create(
            name=task_name,
            type="arc",
            config=task.to_payload().config,
        )
        retrieved = client.tasks.retrieve(key=created.key)
        assert retrieved.key == created.key
        assert retrieved.type == "arc"

    def test_wrap_retrieved_task_as_arc_task(self, client: sy.Synnax):
        """Should wrap a retrieved task as an ArcTask."""
        arc = client.arcs.create(
            name=f"test-arc-task-wrap-{uuid4()}",
            mode="text",
        )
        task_name = f"Test Arc Task {uuid4()}"
        task = sy.arc.Task(name=task_name, arc_key=arc.key)
        created = client.tasks.create(
            name=task_name,
            type="arc",
            config=task.to_payload().config,
        )
        retrieved = client.tasks.retrieve(key=created.key)
        # Wrap the generic Task as an ArcTask
        arc_task = sy.arc.Task(internal=retrieved)
        assert arc_task.config.arc_key == str(arc.key)
        assert arc_task.config.auto_start is False



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import uuid4

import pytest

import synnax as sy


@pytest.mark.rack
class TestRackClient:
    def test_create_single(self, client: sy.Synnax):
        """It should correctly create a single rack"""
        rack = client.racks.create(name="test")
        assert rack.key != 0

    def test_create_multiple(self, client: sy.Synnax):
        """Should create multiple racks"""
        r1 = sy.Rack(name="test1")
        r2 = sy.Rack(name="test2")
        racks = client.racks.create(racks=[r1, r2])
        assert len(racks) == 2
        assert racks[0].name == "test1"
        assert racks[1].name == "test2"

    def test_retrieve_by_name(self, client: sy.Synnax):
        """Should retrieve a rack by name"""
        name = str(uuid4())
        rack = client.racks.create(name=name)
        res = client.racks.retrieve(name=name)
        assert res.name == name
        assert res.key == rack.key

    def test_retrieve_by_key(self, client: sy.Synnax):
        """Should retrieve a rack by key"""
        rack = client.racks.create(name="test")
        res = client.racks.retrieve(key=rack.key)
        assert res.key == rack.key
        assert res.name == rack.name

    def test_retrieve_multiple(self, client: sy.Synnax):
        """Should retrieve multiple racks"""
        r1 = client.racks.create(name="test1")
        r2 = client.racks.create(name="test2")
        racks = client.racks.retrieve(keys=[r1.key, r2.key])
        assert len(racks) == 2
        assert {r.key for r in racks} == {r1.key, r2.key}

    def test_delete(self, client: sy.Synnax):
        """Should delete a rack"""
        rack = client.racks.create(name="test")
        client.racks.delete([rack.key])
        with pytest.raises(sy.NotFoundError):
            client.racks.retrieve(key=rack.key)

    def test_delete_rack_attached(self, client: sy.Synnax):
        """Should raise a validation error if devices are attached to the rack"""
        rack = client.racks.create(name="test")
        client.devices.create(
            key=str(uuid4()), name="test", rack=rack.key, location="dev1"
        )
        with pytest.raises(
            sy.ValidationError,
            match="cannot delete rack when devices are still attached",
        ):
            client.racks.delete([rack.key])

    def test_delete_task_attached(self, client: sy.Synnax):
        """Should raise a validation error if tasks are attached to the rack"""
        rack = client.racks.create(name="test")
        client.tasks.create(name="test", rack=rack.key)
        with pytest.raises(sy.ValidationError, match="tasks are still attached"):
            client.racks.delete([rack.key])

    def test_retrieve_embedded_rack(self, client: sy.Synnax):
        rack = client.racks.retrieve_embedded_rack()
        assert isinstance(rack, sy.Rack)
        # Cache should return the same rack
        cached_rack = client.racks.retrieve_embedded_rack()
        assert rack.key == cached_rack.key



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import synnax as sy
from examples.control.tpc.common import DAQ_TIME, FUEL_PT_1

client = sy.Synnax()

with client.open_streamer([FUEL_PT_1, DAQ_TIME]) as streamer:
    above_threshold = None
    for frame in streamer:
        if frame[FUEL_PT_1] > 20 and not above_threshold:
            above_threshold = sy.TimeStamp(frame[DAQ_TIME][-1])
        elif frame[FUEL_PT_1] < 20 and above_threshold:
            client.ranges.create(
                name=f"Fuel Above Threshold - " + str(above_threshold)[11:19],
                time_range=sy.TimeRange(start=above_threshold, end=frame[DAQ_TIME][-1]),
                color="#BADA55",
            )
            above_threshold = None



#  Copyright 2026 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import sys

import synnax as sy

if len(sys.argv) < 2:
    print("Usage: python listen_to_channels.py <channel_name> [<channel_name> ...]")
    sys.exit(1)

channel_names = sys.argv[1:]
names_set = set(channel_names)
multiple_channels = len(channel_names) > 1

client = sy.Synnax()

with client.open_streamer(channel_names) as streamer:
    for frame in streamer:
        for channel in frame.channels:
            if channel not in names_set:
                continue
            for value in frame[channel]:
                if multiple_channels:
                    print(f"{channel}: {value}")
                else:
                    print(value)
